from pathlib import Path

import zarr
import numcodecs


if hasattr(numcodecs, 'blosc'):

    numcodecs.blosc.use_threads = False

    compressor = numcodecs.Blosc(cname='zstd',
                                 clevel=2,
                                 shuffle=numcodecs.Blosc.BITSHUFFLE)


def to_zarr(filename, data, window, chunks, root=None):

    """
    Writes data to a zarr file

    Args:
        filename (str): The output file name.
        data (ndarray): The data to write.
        window (namedtuple): A ``rasterio.window.Window`` object.
        chunks (int or tuple): The ``zarr`` chunks.
        root (Optional[object]): The ``zarr`` root.

    Returns:
        ``str``
    """

    p = Path(filename)

    f_base = p.name.split('.')[0]
    d_name = p.parent
    sub_dir = d_name.joinpath('sub_tmp_')
    zarr_file = sub_dir.joinpath('data.zarr').as_posix()

    # sub_dir.mkdir(parents=True, exist_ok=True)

    if not root:
        root = zarr.open(zarr_file, mode='r+')

    group_name = '{BASE}_y{Y:09d}_x{X:09d}_h{H:09d}_w{W:09d}'.format(BASE=f_base,
                                                                     Y=window.row_off,
                                                                     X=window.col_off,
                                                                     H=window.height,
                                                                     W=window.width)

    group = root.create_group(group_name)

    synchronizer = zarr.ProcessSynchronizer('data.sync')

    z = group.array('data',
                    data,
                    compressor=compressor,
                    dtype=data.dtype.name,
                    chunks=chunks,
                    synchronizer=synchronizer)

    group.attrs['row_off'] = window.row_off
    group.attrs['col_off'] = window.col_off
    group.attrs['height'] = window.height
    group.attrs['width'] = window.width

    return zarr_file



from ..errors import logger


class PropertyMixin(object):

    @staticmethod
    def check_sensor(data, sensor=None, return_error=True):

        """
        Checks if a sensor name is provided
        """

        if not sensor:

            if data.gw.sensor:
                sensor = data.gw.sensor
            else:

                if return_error:
                    logger.exception('  A sensor must be provided.')

        return sensor

    @staticmethod
    def check_sensor_band_names(data, sensor, band_names):

        """
        Checks if band names can be collected from a sensor's wavelength names
        """

        if not band_names:

            if isinstance(sensor, str):
                band_names = list(data.gw.wavelengths[sensor]._fields)
            elif isinstance(data.gw.sensor, str):
                band_names = list(data.gw.wavelengths[data.gw.sensor]._fields)

        if not band_names:
            band_names = list(range(1, data.gw.nbands+1))

        return band_names



from . import config
from .core.api import open
from .core import extract
from .core import sample
from .core import subset
from .core import clip
from .core import mask
from .core import coregister
from .core import polygons_to_points
from .core import apply
from .core import to_crs
from .core import transform_crs
from .core import to_raster
from .core import to_vrt
from .core import to_geodataframe
from .core import geodataframe_to_array
from .core import array_to_polygon
from .core import polygon_to_array
from .core import moving
from .core import norm_diff
from .core import evi
from .core import evi2
from .core import nbr
from .core import ndvi
from .core import wi
from .core import tasseled_cap
from .core import coords_to_indices
from .core import indices_to_coords

from .version import __version__

__all__ = ['config',
           'open',
           'extract',
           'sample',
           'subset',
           'clip',
           'mask',
           'coregister',
           'polygons_to_points',
           'apply',
           'to_crs',
           'transform_crs',
           'to_raster',
           'to_vrt',
           'to_geodataframe',
           'geodataframe_to_array',
           'array_to_polygon',
           'polygon_to_array',
           'moving',
           'norm_diff',
           'evi',
           'evi2',
           'nbr',
           'ndvi',
           'wi',
           'tasseled_cap',
           'coords_to_indices',
           'indices_to_coords',
           '__version__']



from ..errors import logger
from .base import PropertyMixin as _PropertyMixin

import numpy as np
import xarray as xr
import dask.array as da


def _create_nodata_array(data, nodata, band_name, var_name):

    """
    Creates a 'no data' Xarray

    Args:
        data (Xarray)
    """

    return xr.DataArray(data=da.zeros((1, data.gw.nrows, data.gw.ncols),
                                      chunks=(1, data.gw.row_chunks, data.gw.col_chunks),
                                      dtype=data.dtype.name) + nodata,
                        coords={band_name: [var_name],
                                'y': data.y.values,
                                'x': data.x.values},
                        dims=(band_name, 'y', 'x'),
                        attrs=data.attrs)


class BandMath(object):

    @staticmethod
    def scale_and_assign(data, band_variable, scale_factor, names, new_names):

        attrs = data.attrs

        if band_variable == 'wavelength':
            band_data = data.sel(wavelength=names) * scale_factor
        else:
            band_data = data.sel(band=names) * scale_factor

        band_data = band_data.assign_coords(coords={band_variable: new_names})
        band_data = band_data.assign_attrs(**attrs)

        return band_data

    @staticmethod
    def mask_and_assign(data,
                        result,
                        band_variable,
                        band_name,
                        nodata,
                        new_name,
                        mask,
                        clip_min,
                        clip_max,
                        scale_factor,
                        sensor):

        """
        Masks a DataArray

        Args:
            data (DataArray or Dataset)
            result (DataArray)
            band_variable (str)
            band_name (str)
            nodata (int or float)
            new_name (str)
            mask (bool)
            clip_min (int)
            clip_max (int)
            scale_factor (float)
            sensor (str)

        Returns:
            ``xarray.DataArray``
        """

        if isinstance(nodata, int) or isinstance(nodata, float):

            if band_variable == 'wavelength':
                result = xr.where(data.sel(wavelength=band_name) == nodata, nodata, result).astype('float64')
            else:
                result = xr.where(data.sel(band=band_name) == nodata, nodata, result).astype('float64')

        if mask:

            if isinstance(data, xr.Dataset):
                result = result.where(data['mask'] < 3)
            else:

                if band_variable == 'wavelength':
                    result = result.where(data.sel(wavelength='mask') < 3)
                else:
                    result = result.where(data.sel(band='mask') < 3)

        new_attrs = data.attrs
        new_attrs['pre-scaling'] = scale_factor
        new_attrs['sensor'] = sensor
        new_attrs['drange'] = (clip_min, clip_max)

        result.clip(min=clip_min, max=clip_max)

        if 'time' in result.coords:

            if band_variable == 'wavelength':
                result = result.assign_coords(wavelength=new_name)
            else:
                result = result.assign_coords(band=new_name)

            result = result.expand_dims(dim=band_variable)

            # Ensure expected order
            result = result.transpose('time', 'band', 'y', 'x')

        else:

            result = result.assign_coords(coords={band_variable: new_name})
            result = result.expand_dims(dim='band')

        result = result.assign_attrs(**new_attrs)

        return result

    def norm_diff_math(self, data, b1, b2, name, sensor, nodata=None, mask=False, scale_factor=1.0):

        """
        Normalized difference index --> (b2 - b1) / (b2 + b1)

        Args:
            data (DataArray)
            b1 (DataArray): Band 1
            b2 (DataArray): Band 2
            name (str)
            sensor (str)
            nodata (Optional[int])
            mask (Optional[bool])
            scale_factor (Optional[float])

        Returns:
            ``xarray.DataArray``
        """

        band_variable = 'wavelength' if 'wavelength' in data.coords else 'band'

        band_data = self.scale_and_assign(data, band_variable, scale_factor, [b1, b2], [b1, b2])

        if band_variable == 'wavelength':

            result = ((band_data.sel(wavelength=b2) - band_data.sel(wavelength=b1)) /
                      (band_data.sel(wavelength=b2) + band_data.sel(wavelength=b1))).fillna(nodata).astype('float64')

        else:

            result = ((band_data.sel(band=b2) - band_data.sel(band=b1)) /
                      (band_data.sel(band=b2) + band_data.sel(band=b1))).fillna(nodata).astype('float64')

        return self.mask_and_assign(band_data, result, band_variable, b2, nodata, name, mask, -1, 1, scale_factor, sensor)

    def evi_math(self, data, sensor, wavelengths, nodata=None, mask=False, scale_factor=1.0):

        """
        Enhanced vegetation index

        Returns:
            ``xarray.DataArray``
        """

        l = 1.0
        c1 = 6.0
        c2 = 7.5
        g = 2.5

        band_variable = 'wavelength' if 'wavelength' in data.coords else 'band'

        if 'nir' in data.coords[band_variable].values.tolist():
            nir = 'nir'
            red = 'red'
            blue = 'blue'
        else:
            nir = wavelengths[sensor].nir
            red = wavelengths[sensor].red
            blue = wavelengths[sensor].blue

        data = self.scale_and_assign(data, band_variable, scale_factor, [nir, red, blue], ['nir', 'red', 'blue'])

        if band_variable == 'wavelength':

            result = (g * (data.sel(wavelength='nir') - data.sel(wavelength='red')) /
                      (data.sel(wavelength='nir') * c1 * data.sel(wavelength='red') - c2 * data.sel(wavelength='blue') + l)).fillna(nodata).astype('float64')

        else:

            result = (g * (data.sel(band='nir') - data.sel(band='red')) /
                      (data.sel(band='nir') * c1 * data.sel(band='red') - c2 * data.sel(band='blue') + l)).fillna(nodata).astype('float64')

        return self.mask_and_assign(data, result, band_variable, 'nir', nodata, 'evi', mask, 0, 1, scale_factor, sensor)

    def evi2_math(self, data, sensor, wavelengths, nodata=None, mask=False, scale_factor=1.0):

        """
        Two-band enhanced vegetation index

        Returns:
            ``xarray.DataArray``
        """

        band_variable = 'wavelength' if 'wavelength' in data.coords else 'band'

        if 'nir' in data.coords[band_variable].values.tolist():
            nir = 'nir'
            red = 'red'
        else:
            nir = wavelengths[sensor].nir
            red = wavelengths[sensor].red

        data = self.scale_and_assign(data, band_variable, scale_factor, [nir, red], ['nir', 'red'])

        if band_variable == 'wavelength':

            result = (2.5 * ((data.sel(wavelength='nir') - data.sel(wavelength='red')) /
                             (data.sel(wavelength='nir') + 1.0 + (2.4 * (data.sel(wavelength='red')))))).fillna(nodata).astype('float64')

        else:

            result = (2.5 * ((data.sel(band='nir') - data.sel(band='red')) /
                             (data.sel(band='nir') + 1.0 + (2.4 * (data.sel(band='red')))))).fillna(nodata).astype('float64')

        return self.mask_and_assign(data, result, band_variable, 'nir', nodata, 'evi2', mask, 0, 1, scale_factor, sensor)

    def nbr_math(self, data, sensor, wavelengths, nodata=None, mask=False, scale_factor=1.0):

        """
        Normalized burn ratio

        Returns:
            ``xarray.DataArray``
        """

        band_variable = 'wavelength' if 'wavelength' in data.coords else 'band'

        if 'nir' in data.coords[band_variable].values.tolist():
            nir = 'nir'
            swir2 = 'swir2'
        else:
            nir = wavelengths[sensor].nir
            swir2 = wavelengths[sensor].swir2

        return self.norm_diff_math(data, swir2, nir, 'nbr', sensor, nodata=nodata, mask=mask, scale_factor=scale_factor)

    def ndvi_math(self, data, sensor, wavelengths, nodata=None, mask=False, scale_factor=1.0):

        """
        Normalized difference vegetation index

        Returns:
            ``xarray.DataArray``
        """

        band_variable = 'wavelength' if 'wavelength' in data.coords else 'band'

        if 'nir' in data.coords[band_variable].values.tolist():
            nir = 'nir'
            red = 'red'
        else:
            nir = wavelengths[sensor].nir
            red = wavelengths[sensor].red

        return self.norm_diff_math(data, red, nir, 'ndvi', sensor, nodata=nodata, mask=mask, scale_factor=scale_factor)

    def wi_math(self, data, sensor, wavelengths, nodata=None, mask=False, scale_factor=1.0):

        """
        Woody index

        Returns:
            ``xarray.DataArray``
        """

        band_variable = 'wavelength' if 'wavelength' in data.coords else 'band'

        if 'swir1' in data.coords[band_variable].values.tolist():
            swir1 = 'swir1'
            red = 'red'
        else:
            swir1 = wavelengths[sensor].swir1
            red = wavelengths[sensor].red

        data = self.scale_and_assign(data, band_variable, scale_factor, [swir1, red], ['swir1', 'red'])

        if band_variable == 'wavelength':
            result = data.sel(wavelength='swir1') + data.sel(wavelength='red')
        else:
            result = data.sel(band='swir1') + data.sel(band='red')

        result = result.where(result > 0.5, 0, 1.0 - (result / 0.5)).fillna(nodata).astype('float64')

        return self.mask_and_assign(data, result, band_variable, 'red', nodata, 'wi', mask, 0, 1, scale_factor, sensor)


def linear_transform(data, bands, scale, offset):

    r"""
    Linearly scales bands using a scale and an offset

    Args:
        data (DataArray): The ``xarray.DataArray`` to transform.
        bands (1d array-like): The list of bands to transform.
        scale (1d array): The scale coefficients.
        offset (1d array): The offset coefficients.

    Equation:

        .. math::
            y = scale \times band + offset

    Returns:
        ``xarray.DataArray``
    """

    scalexr = xr.DataArray(scale,
                           coords=[bands],
                           dims=['band'])

    offsetxr = xr.DataArray(offset,
                            coords=[bands],
                            dims=['band'])

    return data * scalexr + offsetxr


class TasseledCapLookup(object):

    @staticmethod
    def get_coefficients(wavelengths, sensor):

        lookup_dict = dict(aster=np.array([[0.3909, -0.0318, 0.4571],
                                           [0.5224, -0.1031, 0.4262],
                                           [0.1184, 0.9422, -0.1568],
                                           [0.3233, 0.2512, 0.2809],
                                           [0.305, -0.0737, -0.2417],
                                           [0.3571, -0.069, -0.3269],
                                           [0.3347, -0.0957, -0.4077],
                                           [0.3169, -0.1195, -0.3731],
                                           [0.151, -0.0625, -0.1877]], dtype='float64'),
                           cbers2=np.array([[0.509, -0.494, 0.581],
                                            [0.431, -0.318, -0.07],
                                            [0.33, -0.324, -0.811],
                                            [0.668, 0.741, 0.003]], dtype='float64'),
                           ik=np.array([[0.326, -0.311, -0.612],
                                        [0.509, -0.256, -0.312],
                                        [0.56, -0.325, 0.722],
                                        [0.567, 0.819, -0.081]], dtype='float64'),
                           l4=np.array([[0.433, -0.29, -0.829],
                                        [0.632, -0.562, 0.522],
                                        [0.586, 0.6, -0.039],
                                        [0.264, 0.491, 0.194]], dtype='float64'),
                           l5=np.array([[0.3037, -0.2848, 0.1509],
                                        [0.2793, -0.2435, 0.1793],
                                        [0.4343, -0.5436, 0.3299],
                                        [0.5585, 0.7243, 0.3406],
                                        [0.5082, 0.084, -0.7112],
                                        [0.1863, -0.18, -0.4572]], dtype='float64'),
                           l7=np.array([[0.3561, -0.3344, 0.2626],
                                        [0.3972, -0.3544, 0.2141],
                                        [0.3904, -0.4556, 0.0926],
                                        [0.6966, 0.6966, 0.0656],
                                        [0.2286, -0.0242, -0.7629],
                                        [0.1596, -0.263, -0.5388]], dtype='float64'),
                           l8=np.array([[0.3029, -0.2941, 0.1511],
                                        [0.2786, -0.243, 0.1973],
                                        [0.4733, -0.5424, 0.3283],
                                        [0.5599, 0.7276, 0.3407],
                                        [0.508, 0.0713, -0.7117],
                                        [0.1872, -0.1608, -0.4559]], dtype='float64'),
                           modis=np.array([[0.4395, -0.4064, 0.1147],
                                           [0.5945, 0.5129, 0.2489],
                                           [0.2460, -0.2744, 0.2408],
                                           [0.3918, -0.2893, 0.3132],
                                           [0.3506, 0.4882, -0.3122],
                                           [0.2136, -0.0036, -0.6416],
                                           [0.2678, -0.4169, -0.5087]], dtype='float64'),
                           qb=np.array([[0.319, -0.121, 0.652],
                                        [0.542, -0.331, 0.375],
                                        [0.49, -0.517, -0.639],
                                        [0.604, 0.78, -0.163]], dtype='float64'),
                           rapideye=np.array([[-0.293, -0.406, 0.572],
                                              [-0.354, -0.367, 0.402],
                                              [-0.372, -0.446, -0.494],
                                              [-0.44, -0.128, -0.498],
                                              [-0.676, 0.697, 0.138]], dtype='float64'))

        return xr.DataArray(data=lookup_dict[sensor],
                            coords={'coeff': ['brightness', 'greenness', 'wetness'],
                                    'band': list(wavelengths[sensor]._fields)},
                            dims=('band', 'coeff'))


class TasseledCap(_PropertyMixin, TasseledCapLookup):

    def tasseled_cap(self, data, nodata=None, sensor=None, scale_factor=1.0):

        r"""
        Applies a tasseled cap transformation

        Args:
            data (DataArray): The ``xarray.DataArray`` to process.
            nodata (Optional[int or float]): A 'no data' value to fill NAs with.
            sensor (Optional[str]): The data's sensor.
            scale_factor (Optional[float]): A scale factor to apply to the data.

        Examples:
            >>> import geowombat as gw
            >>>
            >>> with gw.config.update(sensor='qb', scale_factor=0.0001):
            >>>     with gw.open('image.tif', band_names=['blue', 'green', 'red', 'nir']) as ds:
            >>>         tcap = gw.tasseled_cap(ds)

        Returns:
            ``xarray.DataArray``

        References:

            ASTER:
                See :cite:`yajuan_danfeng_2005`

            CBERS2:
                See :cite:`sheng_etal_2011`

            IKONOS:
                See :cite:`pereira_2006`

            Landsat ETM+:
                See :cite:`huang_etal_2002`

            Landsat OLI:
                See :cite:`baig_etal_2014`

            MODIS:
                See :cite:`lobser_cohen_2007`

            Quickbird:
                See :cite:`yarbrough_etal_2005`

            RapidEye:
                See :cite:`arnett_etal_2014`
        """

        sensor = self.check_sensor(data, sensor)

        if not isinstance(nodata, int) and not isinstance(nodata, float):
            nodata = data.gw.nodata

        if scale_factor == 1.0:
            scale_factor = data.gw.scale_factor

        tc_coefficients = self.get_coefficients(data.gw.wavelengths, sensor)

        tcap = ((data * scale_factor) * tc_coefficients).sum(dim='band').fillna(nodata).transpose('coeff', 'y', 'x').rename({'coeff': 'band'})

        tcap.attrs = data.attrs

        return tcap


class VegetationIndices(_PropertyMixin, BandMath):

    def norm_diff(self, data, b1, b2, sensor=None, nodata=None, mask=False, scale_factor=1.0):

        r"""
        Calculates the normalized difference band ratio

        Args:
            data (DataArray): The ``xarray.DataArray`` to process.
            b1 (str): The band name of the first band.
            b2 (str): The band name of the second band.
            sensor (Optional[str]): sensor (Optional[str]): The data's sensor.
            nodata (Optional[int or float]): A 'no data' value to fill NAs with.
            mask (Optional[bool]): Whether to mask the results.
            scale_factor (Optional[float]): A scale factor to apply to the data.

        Equation:

            .. math::
                {norm}_{diff} = \frac{b2 - b1}{b2 + b1}

        Returns:

            ``xarray.DataArray``:

                Data range: -1 to 1
        """

        sensor = self.check_sensor(data, sensor)

        if not isinstance(nodata, int) and not isinstance(nodata, float):
            nodata = data.gw.nodata

        if scale_factor == 1.0:
            scale_factor = data.gw.scale_factor

        return self.norm_diff_math(data, b1, b2, 'norm-diff', sensor, nodata=nodata, mask=mask, scale_factor=scale_factor)

    def evi(self, data, nodata=None, mask=False, sensor=None, scale_factor=1.0):

        r"""
        Calculates the enhanced vegetation index

        Args:
            data (DataArray): The ``xarray.DataArray`` to process.
            nodata (Optional[int or float]): A 'no data' value to fill NAs with.
            mask (Optional[bool]): Whether to mask the results.
            sensor (Optional[str]): The data's sensor.
            scale_factor (Optional[float]): A scale factor to apply to the data.

        Equation:

            .. math::

                EVI = 2.5 \times \frac{NIR - red}{NIR \times 6 \times red - 7.5 \times blue + 1}

        Returns:

            ``xarray.DataArray``:

                Data range: 0 to 1
        """

        sensor = self.check_sensor(data, sensor)

        if not isinstance(nodata, int) and not isinstance(nodata, float):
            nodata = data.gw.nodata

        if scale_factor == 1.0:
            scale_factor = data.gw.scale_factor

        return self.evi_math(data, sensor, data.gw.wavelengths, nodata=nodata, mask=mask, scale_factor=scale_factor)

    def evi2(self, data, nodata=None, mask=False, sensor=None, scale_factor=1.0):

        r"""
        Calculates the two-band modified enhanced vegetation index

        Args:
            data (DataArray): The ``xarray.DataArray`` to process.
            nodata (Optional[int or float]): A 'no data' value to fill NAs with.
            mask (Optional[bool]): Whether to mask the results.
            sensor (Optional[str]): The data's sensor.
            scale_factor (Optional[float]): A scale factor to apply to the data.

        Equation:

            .. math::

                EVI2 = 2.5 \times \frac{NIR - red}{NIR + 1 + 2.4 \times red}

        Returns:

            ``xarray.DataArray``:

                Data range: 0 to 1
        """

        sensor = self.check_sensor(data, sensor)

        if not isinstance(nodata, int) and not isinstance(nodata, float):
            nodata = data.gw.nodata

        if scale_factor == 1.0:
            scale_factor = data.gw.scale_factor

        return self.evi2_math(data, sensor, data.gw.wavelengths, nodata=nodata, mask=mask, scale_factor=scale_factor)

    def nbr(self, data, nodata=None, mask=False, sensor=None, scale_factor=1.0):

        r"""
        Calculates the normalized burn ratio

        Args:
            data (DataArray): The ``xarray.DataArray`` to process.
            nodata (Optional[int or float]): A 'no data' value to fill NAs with.
            mask (Optional[bool]): Whether to mask the results.
            sensor (Optional[str]): The data's sensor.
            scale_factor (Optional[float]): A scale factor to apply to the data.

        Equation:

            .. math::
                NBR = \frac{NIR - SWIR1}{NIR + SWIR1}

        Returns:
            ``xarray.DataArray``
        """

        sensor = self.check_sensor(data, sensor)

        if not isinstance(nodata, int) and not isinstance(nodata, float):
            nodata = data.gw.nodata

        if scale_factor == 1.0:
            scale_factor = data.gw.scale_factor

        return self.nbr_math(data, sensor, data.gw.wavelengths, nodata=nodata, mask=mask, scale_factor=scale_factor)

    def ndvi(self, data, nodata=None, mask=False, sensor=None, scale_factor=1.0):

        r"""
        Calculates the normalized difference vegetation index

        Args:
            data (DataArray): The ``xarray.DataArray`` to process.
            nodata (Optional[int or float]): A 'no data' value to fill NAs with.
            mask (Optional[bool]): Whether to mask the results.
            sensor (Optional[str]): The data's sensor.
            scale_factor (Optional[float]): A scale factor to apply to the data.

        Equation:

            .. math::
                NDVI = \frac{NIR - red}{NIR + red}

        Returns:

            ``xarray.DataArray``:

                Data range: -1 to 1
        """

        sensor = self.check_sensor(data, sensor)

        if not isinstance(nodata, int) and not isinstance(nodata, float):
            nodata = data.gw.nodata

        if scale_factor == 1.0:
            scale_factor = data.gw.scale_factor

        return self.ndvi_math(data, sensor, data.gw.wavelengths, nodata=nodata, mask=mask, scale_factor=scale_factor)

    def wi(self, data, nodata=None, mask=False, sensor=None, scale_factor=1.0):

        r"""
        Calculates the woody vegetation index

        Args:
            data (DataArray): The ``xarray.DataArray`` to process.
            nodata (Optional[int or float]): A 'no data' value to fill NAs with.
            mask (Optional[bool]): Whether to mask the results.
            sensor (Optional[str]): The data's sensor.
            scale_factor (Optional[float]): A scale factor to apply to the data.

        Equation:

            .. math::
                WI = SWIR1 + red

        Returns:

            ``xarray.DataArray``:

                Data range: 0 to 1
        """

        sensor = self.check_sensor(data, sensor)

        if not isinstance(nodata, int) and not isinstance(nodata, float):
            nodata = data.gw.nodata

        if scale_factor == 1.0:
            scale_factor = data.gw.scale_factor

        return self.wi_math(data, sensor, data.gw.wavelengths, nodata=nodata, mask=mask, scale_factor=scale_factor)



import os
import multiprocessing as multi

from .util import sample_feature
from .util import lazy_wombat

import numpy as np
import dask.array as da
import xarray as xr
import pandas as pd
import geopandas as gpd
from rasterio.features import rasterize, shapes
from rasterio.warp import aligned_target
import shapely
from shapely.geometry import Polygon
from affine import Affine
from tqdm import tqdm
from deprecated import deprecated

shapely.speedups.enable()


def _iter_func(a):
    return a


class Converters(object):

    @staticmethod
    def indices_to_coords(col_index, row_index, transform):

        """
        Converts array indices to map coordinates

        Args:
            col_index (float or 1d array): The column index.
            row_index (float or 1d array): The row index.
            transform (Affine, DataArray, or tuple): The affine transform.

        Returns:

            ``tuple``:

                (x, y)

        Example:
            >>> import geowombat as gw
            >>> from geowombat.core import indices_to_coords
            >>>
            >>> with gw.open('image.tif') as src:
            >>>     x, y = indices_to_coords(j, i, src)
        """

        if not isinstance(transform, Affine):

            if isinstance(transform, tuple):
                transform = Affine(*transform)
            elif isinstance(transform, xr.DataArray):
                transform = transform.gw.meta.affine
            else:
                logger.exception(
                    '  The transform must be an instance of affine.Affine, an xarray.DataArray, or a tuple')
                raise TypeError

        return transform * (col_index, row_index)

    @staticmethod
    def coords_to_indices(x, y, transform):

        """
        Converts map coordinates to array indices

        Args:
            x (float or 1d array): The x coordinates.
            y (float or 1d array): The y coordinates.
            transform (object): The affine transform.

        Returns:

            ``tuple``:

                (col_index, row_index)

        Example:
            >>> import geowombat as gw
            >>> from geowombat.core import coords_to_indices
            >>>
            >>> with gw.open('image.tif') as src:
            >>>     j, i = coords_to_indices(x, y, src)
        """

        if not isinstance(transform, Affine):

            if isinstance(transform, tuple):
                transform = Affine(*transform)
            elif isinstance(transform, xr.DataArray):
                transform = transform.gw.meta.affine
            else:
                logger.exception(
                    '  The transform must be an instance of affine.Affine, an xarray.DataArray, or a tuple')
                raise TypeError

        col_index, row_index = ~transform * (x, y)

        return np.int64(col_index), np.int64(row_index)

    @staticmethod
    @deprecated('Deprecated since 1.0.6. Use indices_to_coords() instead.')
    def ij_to_xy(j, i, transform):

        """
        .. deprecated:: 1.0.6
            Use :func:`geowombat.indices_to_coords()` instead.

        Converts to array indices to map coordinates

        Args:
            j (float or 1d array): The column index.
            i (float or 1d array): The row index.
            transform (object): The affine transform.

        Returns:
            x, y
        """

        return transform * (j, i)

    @staticmethod
    @deprecated('Deprecated since 1.0.6. Use coords_to_indices() instead.')
    def xy_to_ij(x, y, transform):

        """
        .. deprecated:: 1.0.6
            Use :func:`geowombat.coords_to_indices()` instead.

        Converts map coordinates to array indices

        Args:
            x (float or 1d array): The x coordinates.
            y (float or 1d array): The y coordinates.
            transform (object): The affine transform.

        Returns:
            j, i
        """

        if not isinstance(transform, Affine):
            transform = Affine(*transform)

        x, y = ~transform * (x, y)

        return np.int64(x), np.int64(y)

    @staticmethod
    def dask_to_xarray(data,
                       dask_data,
                       band_names):

        """
        Converts a Dask array to an Xarray DataArray

        Args:
            data (DataArray): The DataArray with attribute information.
            dask_data (Dask Array): The Dask array to convert.
            band_names (1d array-like): The output band names.

        Returns:
            ``xarray.DataArray``
        """

        if len(dask_data.shape) == 2:
            dask_data = dask_data.reshape(1, dask_data.shape[0], dask_data.shape[1])

        return xr.DataArray(dask_data,
                            dims=('band', 'y', 'x'),
                            coords={'band': band_names,
                                    'y': data.y,
                                    'x': data.x},
                            attrs=data.attrs)

    @staticmethod
    def ndarray_to_xarray(data,
                          numpy_data,
                          band_names):

        """
        Converts a NumPy array to an Xarray DataArray

        Args:
            data (DataArray): The DataArray with attribute information.
            numpy_data (Dask Array): The Dask array to convert.
            band_names (1d array-like): The output band names.

        Returns:
            ``xarray.DataArray``
        """

        if len(numpy_data.shape) == 2:
            numpy_data = numpy_data[np.newaxis, :, :]

        return xr.DataArray(da.from_array(numpy_data,
                                          chunks=(1, data.gw.row_chunks, data.gw.col_chunks)),
                            dims=('band', 'y', 'x'),
                            coords={'band': band_names,
                                    'y': data.y,
                                    'x': data.x},
                            attrs=data.attrs)

    @staticmethod
    def xarray_to_xdataset(data_array,
                           band_names,
                           time_names,
                           ycoords=None,
                           xcoords=None,
                           attrs=None):

        """
        Converts an Xarray DataArray to a Xarray Dataset

        Args:
            data_array (DataArray)
            band_names (list)
            time_names (list)
            ycoords (1d array-like)
            xcoords (1d array-like)
            attrs (dict)

        Returns:
            Dataset
        """

        if len(data_array.shape) == 2:
            data_array = data_array.expand_dims('band')

        if len(data_array.shape) == 4:
            n_bands = data_array.shape[1]
        else:
            n_bands = data_array.shape[0]

        if not band_names:

            if n_bands == 1:
                band_names = ['1']
            else:
                band_names = list(map(str, range(1, n_bands + 1)))

        if time_names:

            return xr.Dataset({'bands': (['date', 'band', 'y', 'x'], data_array)},
                              coords={'date': time_names,
                                      'band': band_names,
                                      'y': ('y', ycoords),
                                      'x': ('x', xcoords)},
                              attrs=attrs)

        else:

            return xr.Dataset({'bands': (['band', 'y', 'x'], data_array.data)},
                              coords={'band': band_names,
                                      'y': ('y', data_array.y),
                                      'x': ('x', data_array.x)},
                              attrs=data_array.attrs)

    def prepare_points(self,
                       data,
                       aoi,
                       frac=1.0,
                       all_touched=False,
                       id_column='id',
                       mask=None,
                       n_jobs=8,
                       verbose=0):

        if isinstance(aoi, gpd.GeoDataFrame):
            df = aoi
        else:

            if isinstance(aoi, str):

                if not os.path.isfile(aoi):
                    logger.exception('  The AOI file does not exist.')

                df = gpd.read_file(aoi)

            else:
                logger.exception('  The AOI must be a vector file or a GeoDataFrame.')

        # Re-project the data to match the image CRS
        if isinstance(df.crs, str):

            if df.crs.lower().startswith('+proj'):

                if data.crs != df.crs:
                    df = df.to_crs(data.crs)

        elif isinstance(df.crs, int):

            if data.crs != CRS.from_epsg(df.crs).to_proj4():
                df = df.to_crs(data.crs)

        else:

            if data.crs != CRS.from_dict(df.crs).to_proj4():
                df = df.to_crs(data.crs)

        if verbose > 0:
            logger.info('  Checking geometry validity ...')

        # Ensure all geometry is valid
        df = df[df['geometry'].apply(lambda x_: x_ is not None)]

        if verbose > 0:
            logger.info('  Checking geometry extent ...')

        # Remove data outside of the image bounds
        if type(df.iloc[0].geometry) == Polygon:

            df = gpd.overlay(df,
                             gpd.GeoDataFrame(data=[0],
                                              geometry=[data.gw.meta.geometry],
                                              crs=df.crs),
                             how='intersection')

        else:

            # Clip points to the image bounds
            df = df[df.geometry.intersects(data.gw.unary_union)]

        if isinstance(mask, Polygon) or isinstance(mask, gpd.GeoDataFrame):

            if isinstance(mask, gpd.GeoDataFrame):

                if CRS.from_dict(mask.crs).to_proj4() != CRS.from_dict(df.crs).to_proj4():
                    mask = mask.to_crs(df.crs)

            if verbose > 0:
                logger.info('  Clipping geometry ...')

            df = df[df.within(mask)]

            if df.empty:
                logger.exception('  No geometry intersects the user-provided mask.')

        # Subset the DataArray
        # minx, miny, maxx, maxy = df.total_bounds
        #
        # obj_subset = self._obj.gw.subset(left=float(minx)-self._obj.res[0],
        #                                  top=float(maxy)+self._obj.res[0],
        #                                  right=float(maxx)+self._obj.res[0],
        #                                  bottom=float(miny)-self._obj.res[0])

        # Convert polygons to points
        if type(df.iloc[0].geometry) == Polygon:

            if verbose > 0:
                logger.info('  Converting polygons to points ...')

            df = self.polygons_to_points(data,
                                         df,
                                         frac=frac,
                                         all_touched=all_touched,
                                         id_column=id_column,
                                         n_jobs=n_jobs)

        # Ensure a unique index
        df.index = list(range(0, df.shape[0]))

        return df

    @staticmethod
    def polygons_to_points(data,
                           df,
                           frac=1.0,
                           all_touched=False,
                           id_column='id',
                           n_jobs=1):

        """
        Converts polygons to points

        Args:
            data (DataArray or Dataset): The ``xarray.DataArray`` or ``xarray.Dataset``.
            df (GeoDataFrame): The ``geopandas.GeoDataFrame`` containing the geometry to rasterize.
            frac (Optional[float]): A fractional subset of points to extract in each feature.
            all_touched (Optional[bool]): The ``all_touched`` argument is passed to ``rasterio.features.rasterize``.
            id_column (Optional[str]): The 'id' column.
            n_jobs (Optional[int]): The number of features to rasterize in parallel.

        Returns:
            ``geopandas.GeoDataFrame``
        """

        meta = data.gw.meta

        dataframes = list()

        with multi.Pool(processes=n_jobs) as pool:

            for i in tqdm(pool.imap(_iter_func, range(0, df.shape[0])), total=df.shape[0]):

                # Get the current feature's geometry
                dfrow = df.iloc[i]

                point_df = sample_feature(dfrow[id_column],
                                          dfrow.geometry,
                                          data.crs,
                                          data.res,
                                          all_touched,
                                          meta,
                                          frac)

                if not point_df.empty:
                    dataframes.append(point_df)

        dataframes = pd.concat(dataframes, axis=0)

        # Make the points unique
        dataframes.loc[:, 'point'] = np.arange(0, dataframes.shape[0])

        return dataframes

    @staticmethod
    def array_to_polygon(data, mask=None, connectivity=4, num_workers=1):

        """
        Converts an ``xarray.DataArray` to a ``geopandas.GeoDataFrame``

        Args:
            data (DataArray): The ``xarray.DataArray`` to convert.
            mask (Optional[str, numpy ndarray, or rasterio Band object]): Must evaluate to bool (rasterio.bool_ or rasterio.uint8).
                Values of False or 0 will be excluded from feature generation. Note well that this is the inverse sense from
                Numpy's, where a mask value of True indicates invalid data in an array. If source is a Numpy masked array
                and mask is None, the source's mask will be inverted and used in place of mask. if ``mask`` is equal to
                'source', then ``data`` is used as the mask.
            connectivity (Optional[int]): Use 4 or 8 pixel connectivity for grouping pixels into features.
            num_workers (Optional[int]): The number of parallel workers to send to ``dask.compute``.

        Returns:
            ``geopandas.GeoDataFrame``

        Example:
            >>> import geowombat as gw
            >>>
            >>> with gw.open('image.tif') as src:
            >>>
            >>>     # Convert the input image to a GeoDataFrame
            >>>     df = gw.array_to_polygon(src,
            >>>                              mask='source',
            >>>                              num_workers=8)
        """

        if not hasattr(data, 'transform'):
            logger.exception("  The data should have a 'transform' object.")

        if not hasattr(data, 'crs'):
            logger.exception("  The data should have a 'crs' object.")

        if isinstance(mask, str):

            if mask == 'source':
                mask = data.astype('uint8').data.compute(num_workers=num_workers)

        poly_objects = shapes(data.data.compute(num_workers=num_workers),
                              mask=mask,
                              connectivity=connectivity,
                              transform=data.transform)

        poly_geom = [Polygon(p[0]['coordinates'][0]) for p in poly_objects]

        return gpd.GeoDataFrame(data=np.ones(len(poly_geom), dtype='uint8'),
                                geometry=poly_geom,
                                crs=data.crs)

    @staticmethod
    @deprecated('Deprecated since 1.2.0. Use array_to_polygon() instead.')
    def to_geodataframe(data, mask=None, connectivity=4, num_workers=1):

        """
        .. deprecated:: 1.2.0
            Use :func:`geowombat.array_to_polygon()` instead.

        Converts a Dask array to a GeoDataFrame

        Args:
            data (DataArray): The ``xarray.DataArray`` to convert.
            mask (Optional[str, numpy ndarray, or rasterio Band object]): Must evaluate to bool (rasterio.bool_ or rasterio.uint8).
                Values of False or 0 will be excluded from feature generation. Note well that this is the inverse sense from
                Numpy's, where a mask value of True indicates invalid data in an array. If source is a Numpy masked array
                and mask is None, the source's mask will be inverted and used in place of mask. if ``mask`` is equal to
                'source', then ``data`` is used as the mask.
            connectivity (Optional[int]): Use 4 or 8 pixel connectivity for grouping pixels into features.
            num_workers (Optional[int]): The number of parallel workers to send to ``dask.compute``.

        Returns:
            ``GeoDataFrame``

        Example:
            >>> import geowombat as gw
            >>>
            >>> with gw.open('image.tif') as src:
            >>>
            >>>     # Convert the input image to a GeoDataFrame
            >>>     df = gw.to_geodataframe(src,
            >>>                             mask='source',
            >>>                             num_workers=8)
        """

        if not hasattr(data, 'transform'):
            logger.exception("  The data should have a 'transform' object.")

        if not hasattr(data, 'crs'):
            logger.exception("  The data should have a 'crs' object.")

        if isinstance(mask, str):

            if mask == 'source':
                mask = data.astype('uint8').data.compute(num_workers=num_workers)

        poly_objects = shapes(data.data.compute(num_workers=num_workers),
                              mask=mask,
                              connectivity=connectivity,
                              transform=data.transform)

        poly_geom = [Polygon(p[0]['coordinates'][0]) for p in poly_objects]

        return gpd.GeoDataFrame(data=np.ones(len(poly_geom), dtype='uint8'),
                                geometry=poly_geom,
                                crs=data.crs)

    @lazy_wombat
    def polygon_to_array(self,
                         polygon,
                         data=None,
                         cellx=None,
                         celly=None,
                         band_name=None,
                         row_chunks=512,
                         col_chunks=512,
                         src_res=None,
                         fill=0,
                         default_value=1,
                         all_touched=True,
                         dtype='uint8',
                         sindex=None):

        """
        Converts a polygon geometry to an ``xarray.DataArray``.

        Args:
            polygon (GeoDataFrame | str): The ``geopandas.DataFrame`` or file with polygon geometry.
            data (Optional[DataArray]): An ``xarray.DataArray`` to use as a reference.
            cellx (Optional[float]): The output cell x size.
            celly (Optional[float]): The output cell y size.
            band_name (Optional[list]): The ``xarray.DataArray`` band name.
            row_chunks (Optional[int]): The ``dask`` row chunk size.
            col_chunks (Optional[int]): The ``dask`` column chunk size.
            src_res (Optional[tuple]: A source resolution to align to.
            fill (Optional[int]): The output fill value for ``rasterio.features.rasterize``.
            default_value (Optional[int]): The output default value for ``rasterio.features.rasterize``.
            all_touched (Optional[int]): The ``all_touched`` value for ``rasterio.features.rasterize``.
            dtype (Optional[int]): The output data type for ``rasterio.features.rasterize``.
            sindex (Optional[object]): An instanced of ``geopandas.GeoDataFrame.sindex``.

        Returns:
            ``xarray.DataArray``

        Example:
            >>> import geowombat as gw
            >>> import geopandas as gpd
            >>>
            >>> df = gpd.read_file('polygons.gpkg')
            >>>
            >>> # 100x100 cell size
            >>> data = gw.polygon_to_array(df, 100.0, 100.0)
            >>>
            >>> # Align to an existing image
            >>> with gw.open('image.tif') as src:
            >>>     data = gw.polygon_to_array(df, data=src)
        """

        if not band_name:
            band_name = [1]

        if isinstance(polygon, gpd.GeoDataFrame):
            dataframe = polygon
        else:

            if os.path.isfile(polygon):
                dataframe = gpd.read_file(polygon)
            else:
                logger.exception('  The polygon file does not exists.')
                raise OSError

        if isinstance(data, xr.DataArray):

            if dataframe.crs != data.crs:

                # Transform the geometry
                dataframe = dataframe.to_crs(data.crs)

            if sindex:

                # Get the R-tree spatial index
                sindex = dataframe.sindex

            # Get intersecting features
            int_idx = sorted(list(sindex.intersection(tuple(data.gw.geodataframe.bounds.values.flatten()))))

            if not int_idx:

                return self.dask_to_xarray(data, da.zeros((1, data.gw.nrows, data.gw.ncols),
                                                          chunks=(1, data.gw.row_chunks, data.gw.col_chunks),
                                                          dtype=data.dtype.name), [1])

            # Subset to the intersecting features
            dataframe = dataframe.iloc[int_idx]

            # Clip the geometry
            dataframe = gpd.overlay(dataframe, data.gw.geodataframe, how='intersection')

            if dataframe.empty:

                return self.dask_to_xarray(data, da.zeros((1, data.gw.nrows, data.gw.ncols),
                                                          chunks=(1, data.gw.row_chunks, data.gw.col_chunks),
                                                          dtype=data.dtype.name), [1])

            cellx = data.gw.cellx
            celly = data.gw.celly
            row_chunks = data.gw.row_chunks
            col_chunks = data.gw.col_chunks
            src_res = None

            left, bottom, right, top = data.gw.bounds

            dst_height = data.gw.nrows
            dst_width = data.gw.ncols

            dst_transform = data.transform

        else:

            left, bottom, right, top = dataframe.total_bounds.flatten().tolist()

            dst_height = int((top - bottom) / abs(celly))
            dst_width = int((right - left) / abs(cellx))

            dst_transform = Affine(cellx, 0.0, left, 0.0, -celly, top)

        if src_res:
            dst_transform = aligned_target(dst_transform,
                                           dst_width,
                                           dst_height,
                                           src_res)[0]

            left = dst_transform[2]
            top = dst_transform[5]

            dst_transform = Affine(cellx, 0.0, left, 0.0, -celly, top)

        varray = rasterize(dataframe.geometry.values,
                           out_shape=(dst_height, dst_width),
                           transform=dst_transform,
                           fill=fill,
                           default_value=default_value,
                           all_touched=all_touched,
                           dtype=dtype)

        cellxh = abs(cellx) / 2.0
        cellyh = abs(celly) / 2.0

        xcoords = np.arange(left + cellxh, left + cellxh + dst_width * abs(cellx), cellx)
        ycoords = np.arange(top - cellyh, top - cellyh - dst_height * abs(celly), -celly)

        attrs = {'transform': dst_transform[:6],
                 'crs': dataframe.crs,
                 'res': (cellx, celly),
                 'is_tiled': 1}

        return xr.DataArray(data=da.from_array(varray[np.newaxis, :, :],
                                               chunks=(1, row_chunks, col_chunks)),
                            coords={'band': band_name,
                                    'y': ycoords,
                                    'x': xcoords},
                            dims=('band', 'y', 'x'),
                            attrs=attrs)

    @staticmethod
    @deprecated('Deprecated since 1.2.0. Use polygon_to_array() instead.')
    def geodataframe_to_array(dataframe,
                              data=None,
                              cellx=None,
                              celly=None,
                              band_name=None,
                              row_chunks=512,
                              col_chunks=512,
                              src_res=None,
                              fill=0,
                              default_value=1,
                              all_touched=True,
                              dtype='uint8'):

        """
        .. deprecated:: 1.2.0
            Use :func:`geowombat.polygon_to_array()` instead.

        Converts a polygon ``geopandas.GeoDataFrame`` to an ``xarray.DataArray``.

        Args:
            dataframe (GeoDataFrame): The ``geopandas.DataFrame`` or file with polygon geometries.
            data (Optional[DataArray]): An ``xarray.DataArray`` to use as a reference.
            cellx (Optional[float]): The output cell x size.
            celly (Optional[float]): The output cell y size.
            band_name (Optional[list]): The ``xarray.DataArray`` band name.
            row_chunks (Optional[int]): The ``dask`` row chunk size.
            col_chunks (Optional[int]): The ``dask`` column chunk size.
            src_res (Optional[tuple]: A source resolution to align to.
            fill (Optional[int]): The output fill value for ``rasterio.features.rasterize``.
            default_value (Optional[int]): The output default value for ``rasterio.features.rasterize``.
            all_touched (Optional[int]): The ``all_touched`` value for ``rasterio.features.rasterize``.
            dtype (Optional[int]): The output data type for ``rasterio.features.rasterize``.

        Returns:
            ``xarray.DataArray``

        Example:
            >>> import geowombat as gw
            >>> import geopandas as gpd
            >>>
            >>> df = gpd.read_file('polygons.gpkg')
            >>>
            >>> # 100x100 cell size
            >>> data = gw.geodataframe_to_array(df, 100.0, 100.0)
            >>>
            >>> # Align to an existing image
            >>> with gw.open('image.tif') as src:
            >>>
            >>>     data = gw.geodataframe_to_array(df,
            >>>                                     cellx=src.gw.cellx,
            >>>                                     celly=src.gw.celly,
            >>>                                     row_chunks=src.gw.row_chunks,
            >>>                                     col_chunks=src.gw.col_chunks,
            >>>                                     src_res=src_res)
        """

        if not band_name:
            band_name = [1]

        if not isinstance(dataframe, gpd.GeoDataFrame):
            dataframe = gpd.read_file(dataframe)

        if isinstance(data, xr.DataArray):

            if dataframe.crs != data.crs:
                # Transform the geometry
                dataframe = dataframe.to_crs(data.crs)

            # Get the R-tree spatial index
            sindex = dataframe.sindex

            # Get intersecting features
            int_idx = sorted(list(sindex.intersection(tuple(data.gw.geodataframe.bounds.values.flatten()))))

            if not int_idx:

                return dask_to_xarray(data, da.zeros((1, data.gw.nrows, data.gw.ncols),
                                                     chunks=(1, data.gw.row_chunks, data.gw.col_chunks),
                                                     dtype=data.dtype.name), [1])

            # Subset to the intersecting features
            dataframe = dataframe.iloc[int_idx]

            # Clip the geometry
            dataframe = gpd.overlay(dataframe, data.gw.geodataframe, how='intersection')

            if dataframe.empty:

                return dask_to_xarray(data, da.zeros((1, data.gw.nrows, data.gw.ncols),
                                                     chunks=(1, data.gw.row_chunks, data.gw.col_chunks),
                                                     dtype=data.dtype.name), [1])

            cellx = data.gw.cellx
            celly = data.gw.celly
            row_chunks = data.gw.row_chunks
            col_chunks = data.gw.col_chunks
            src_res = None

            left, bottom, right, top = data.gw.bounds

            dst_height = data.gw.nrows
            dst_width = data.gw.ncols

            dst_transform = data.transform

        else:

            left, bottom, right, top = dataframe.total_bounds.flatten().tolist()

            dst_height = int((top - bottom) / abs(celly))
            dst_width = int((right - left) / abs(cellx))

            dst_transform = Affine(cellx, 0.0, left, 0.0, -celly, top)

        if src_res:
            dst_transform = aligned_target(dst_transform,
                                           dst_width,
                                           dst_height,
                                           src_res)[0]

            left = dst_transform[2]
            top = dst_transform[5]

            dst_transform = Affine(cellx, 0.0, left, 0.0, -celly, top)

        varray = rasterize(dataframe.geometry.values,
                           out_shape=(dst_height, dst_width),
                           transform=dst_transform,
                           fill=fill,
                           default_value=default_value,
                           all_touched=all_touched,
                           dtype=dtype)

        cellxh = abs(cellx) / 2.0
        cellyh = abs(celly) / 2.0

        xcoords = np.arange(left + cellxh, left + cellxh + dst_width * abs(cellx), cellx)
        ycoords = np.arange(top - cellyh, top - cellyh - dst_height * abs(celly), -celly)

        attrs = {'transform': dst_transform[:6],
                 'crs': dataframe.crs,
                 'res': (cellx, celly),
                 'is_tiled': 1}

        return xr.DataArray(data=da.from_array(varray[np.newaxis, :, :],
                                               chunks=(1, row_chunks, col_chunks)),
                            coords={'band': band_name,
                                    'y': ycoords,
                                    'x': xcoords},
                            dims=('band', 'y', 'x'),
                            attrs=attrs)


@deprecated('Deprecated since 1.2.0. Use geowombat.core.dask_to_xarray() instead.')
def dask_to_xarray(data, dask_data, band_names):

    """
    .. deprecated:: 1.2.0
        Use :func:`geowombat.core.dask_to_xarray()` instead.

    Converts a Dask array to an Xarray DataArray

    Args:
        data (DataArray): The DataArray with attribute information.
        dask_data (Dask Array): The Dask array to convert.
        band_names (1d array-like): The output band names.

    Returns:
        ``xarray.DataArray``
    """

    if len(dask_data.shape) == 2:
        dask_data = dask_data.reshape(1, dask_data.shape[0], dask_data.shape[1])

    return xr.DataArray(dask_data,
                        dims=('band', 'y', 'x'),
                        coords={'band': band_names,
                                'y': data.y,
                                'x': data.x},
                        attrs=data.attrs)


@deprecated('Deprecated since 1.2.0. Use geowombat.core.ndarray_to_xarray() instead.')
def ndarray_to_xarray(data, numpy_data, band_names):

    """
    .. deprecated:: 1.2.0
        Use :func:`geowombat.core.ndarray_to_xarray()` instead.

    Converts a NumPy array to an Xarray DataArray

    Args:
        data (DataArray): The DataArray with attribute information.
        numpy_data (Dask Array): The Dask array to convert.
        band_names (1d array-like): The output band names.

    Returns:
        ``xarray.DataArray``
    """

    if len(numpy_data.shape) == 2:
        numpy_data = numpy_data[np.newaxis, :, :]

    return xr.DataArray(da.from_array(numpy_data,
                                      chunks=(1, data.gw.row_chunks, data.gw.col_chunks)),
                        dims=('band', 'y', 'x'),
                        coords={'band': band_names,
                                'y': data.y,
                                'x': data.x},
                        attrs=data.attrs)


@deprecated('Deprecated since 1.2.0.')
def xarray_to_xdataset(data_array, band_names, time_names, ycoords=None, xcoords=None, attrs=None):

    """
    .. deprecated:: 1.2.0

    Converts an Xarray DataArray to a Xarray Dataset

    Args:
        data_array (DataArray)
        band_names (list)
        time_names (list)
        ycoords (1d array-like)
        xcoords (1d array-like)
        attrs (dict)

    Returns:
        Dataset
    """

    if len(data_array.shape) == 2:
        data_array = data_array.expand_dims('band')

    if len(data_array.shape) == 4:
        n_bands = data_array.shape[1]
    else:
        n_bands = data_array.shape[0]

    if not band_names:

        if n_bands == 1:
            band_names = ['1']
        else:
            band_names = list(map(str, range(1, n_bands+1)))

    if time_names:

        return xr.Dataset({'bands': (['date', 'band', 'y', 'x'], data_array)},
                             coords={'date': time_names,
                                     'band': band_names,
                                     'y': ('y', ycoords),
                                     'x': ('x', xcoords)},
                             attrs=attrs)

    else:

        return xr.Dataset({'bands': (['band', 'y', 'x'], data_array.data)},
                          coords={'band': band_names,
                                  'y': ('y', data_array.y),
                                  'x': ('x', data_array.x)},
                          attrs=data_array.attrs)



from ..config import config

from . import to_raster, to_vrt, array_to_polygon, moving, extract, sample, subset, clip, mask
from . import norm_diff as gw_norm_diff
from . import evi as gw_evi
from . import evi2 as gw_evi2
from . import nbr as gw_nbr
from . import ndvi as gw_ndvi
from . import wi as gw_wi
from . import tasseled_cap as gw_tasseled_cap
from . import to_crs as _to_crs
from . import transform_crs as _transform_crs
from .properties import DataProperties as _DataProperties
from .util import project_coords
from ..backends import Cluster as _Cluster
from ..util import imshow as gw_imshow
from ..radiometry import BRDF as _BRDF

import numpy as np
import xarray as xr
import joblib


class _UpdateConfig(object):

    def _update_attrs(self):

        if self.config:

            for k, v in self.config.items():
                setattr(self, k, v)

    def _update_kwargs(self, **kwargs):

        if self.config:

            for k, v in self.config.items():

                # rasterio.write keyword arguments
                if k not in kwargs:
                    kwargs[k] = v

        return kwargs


@xr.register_dataset_accessor('gw')
class GeoWombatAccessor(_UpdateConfig, _DataProperties):

    def __init__(self, xarray_obj):

        self._obj = xarray_obj
        self.sensor = None
        self.ax = None

        self.config = config

        self._update_attrs()

    def imshow(self,
               variable='bands',
               band_names=None,
               mask=False,
               nodata=0,
               flip=False,
               text_color='black',
               rot=30,
               **kwargs):

        """
        Shows an image on a plot

        Args:
            variable (Optional[str]): The ``Dataset`` variable to write.
            band_names (Optional[list or str]): The band name or list of band names to plot.
            mask (Optional[bool]): Whether to mask 'no data' values (given by ``nodata``).
            nodata (Optional[int or float]): The 'no data' value.
            flip (Optional[bool]): Whether to flip an RGB array's band order.
            text_color (Optional[str]): The text color.
            rot (Optional[int]): The degree rotation for the x-axis tick labels.
            kwargs (Optional[dict]): Keyword arguments passed to ``xarray.plot.imshow``.

        Returns:
            None

        Examples:
            >>> with gw.open('image.tif', return_as='dataset') as ds:
            >>>     ds.gw.imshow(band_names=['red', 'green', 'red'], mask=True, vmin=0.1, vmax=0.9, robust=True)
        """

        gw_imshow(self._obj[variable],
                  band_names=band_names,
                  mask=mask,
                  nodata=nodata,
                  flip=flip,
                  text_color=text_color,
                  rot=rot,
                  **kwargs)

    def to_raster(self,
                  filename,
                  variable='bands',
                  verbose=0,
                  overwrite=False,
                  driver='GTiff',
                  gdal_cache=512,
                  dtype=None,
                  nodata=None,
                  tags=None,
                  **kwargs):

        """
        Writes an Xarray Dataset to a raster file

        Args:
            filename (str): The output file name to write to.
            variable (Optional[str]): The ``Dataset`` variable to write.
            filename (str): The output file name to write to.
            n_jobs (Optional[str]): The number of parallel chunks to write.
            verbose (Optional[int]): The verbosity level.
            overwrite (Optional[bool]): Whether to overwrite an existing file.
            driver (Optional[str]): The raster driver.
            gdal_cache (Optional[int]): The ``GDAL`` cache size (in MB).
            dtype (Optional[int]): The output data type.
            row_chunks (Optional[int]): The processing row chunk size.
            col_chunks (Optional[int]): The processing column chunk size.
            pool_chunksize (Optional[int]): The `multiprocessing.Pool` chunk size.
            nodata (Optional[int]): A 'no data' value.
            tags (Optional[dict]): Image tags to write to file.
            kwargs (Optional[dict]): Additional keyword arguments to pass to ``rasterio.write``.

        Returns:
            None

        Examples:
            >>> import geowombat as gw
            >>> from geowombat.backends import Cluster
            >>>
            >>> cluster = Cluster(n_workers=4,
            >>>                   threads_per_worker=2,
            >>>                   scheduler_port=0,
            >>>                   processes=False)
            >>>
            >>> cluster.start()
            >>>
            >>> with gw.open('input.tif') as ds:
            >>>     ds.gw.to_raster('output.tif', n_jobs=8)
            >>>
            >>> cluster.stop()
        """

        if not hasattr(self._obj, 'crs'):
            raise AttributeError('The Dataset does not have a `crs` attribute.')

        if not hasattr(self._obj, 'transform'):
            raise AttributeError('The Dataset does not have a `transform` attribute.')

        kwargs = self._update_kwargs(nodata=nodata,
                                     driver=driver,
                                     **kwargs)

        to_raster(self._obj[variable],
                  filename,
                  gdal_cache=gdal_cache,
                  verbose=verbose,
                  overwrite=overwrite,
                  crs=self._obj.crs,
                  transform=self._obj.transform,
                  dtype=dtype,
                  tags=tags,
                  **kwargs)

    def moving(self,
               variable='bands',
               band_coords='band',
               stat='mean',
               perc=50.0,
               w=3,
               n_jobs=1):

        """
        Applies a moving window function to the DataArray

        Args:
            variable (Optional[str]): The variable to compute.
            band_coords (Optional[str]): The band coordinate name.
            stat (Optional[str]): The statistic to compute. Choices are ['mean', 'std', 'var', 'min', 'max'].
            perc (Optional[float]): The percentile to return if ``stat`` = 'perc'.
            w (Optional[int]): The moving window size.
            n_jobs (Optional[int]): The number of bands to process in parallel.

        Returns:
            DataArray
        """

        return moving(self._obj[variable],
                      band_names=self._obj.coords[band_coords].values,
                      w=w,
                      perc=perc,
                      stat=stat,
                      n_jobs=n_jobs)

    def norm_diff(self, b1, b2, variable='bands', nodata=None, mask=False, sensor=None, scale_factor=1.0):
        return gw_norm_diff(self._obj[variable], b1, b2, sensor=sensor, nodata=nodata, mask=mask, scale_factor=scale_factor)

    def evi(self, variable='bands', nodata=None, mask=False, sensor=None, scale_factor=1.0):
        return gw_evi(self._obj[variable], nodata=nodata, mask=mask, sensor=sensor, scale_factor=scale_factor)

    def evi2(self, variable='bands', nodata=None, mask=False, sensor=None, scale_factor=1.0):
        return gw_evi2(self._obj[variable], nodata=nodata, mask=mask, sensor=sensor, scale_factor=scale_factor)

    def nbr(self, variable='bands', nodata=None, mask=False, sensor=None, scale_factor=1.0):
        return gw_nbr(self._obj[variable], nodata=nodata, mask=mask, sensor=sensor, scale_factor=scale_factor)

    def ndvi(self, variable='bands', nodata=None, mask=False, sensor=None, scale_factor=1.0):
        return gw_ndvi(self._obj[variable], nodata=nodata, mask=mask, sensor=sensor, scale_factor=scale_factor)

    def wi(self, variable='bands', nodata=None, mask=False, sensor=None, scale_factor=1.0):
        return gw_wi(self._obj[variable], nodata=nodata, mask=mask, sensor=sensor, scale_factor=scale_factor)

    def tasseled_cap(self, variable='bands', nodata=None, sensor=None, scale_factor=1.0):
        return gw_tasseled_cap(self._obj[variable], nodata=nodata, sensor=sensor, scale_factor=scale_factor)


@xr.register_dataarray_accessor('gw')
class GeoWombatAccessor(_UpdateConfig, _DataProperties):

    """
    Xarray IO class
    """

    def __init__(self, xarray_obj):

        self._obj = xarray_obj
        self.config = config

        self._update_attrs()

    def imshow(self,
               mask=False,
               nodata=0,
               flip=False,
               text_color='black',
               rot=30,
               **kwargs):

        """
        Shows an image on a plot

        Args:
            mask (Optional[bool]): Whether to mask 'no data' values (given by ``nodata``).
            nodata (Optional[int or float]): The 'no data' value.
            flip (Optional[bool]): Whether to flip an RGB array's band order.
            text_color (Optional[str]): The text color.
            rot (Optional[int]): The degree rotation for the x-axis tick labels.
            kwargs (Optional[dict]): Keyword arguments passed to ``xarray.plot.imshow``.

        Returns:
            None

        Examples:
            >>> with gw.open('image.tif') as ds:
            >>>     ds.gw.imshow(band_names=['red', 'green', 'red'], mask=True, vmin=0.1, vmax=0.9, robust=True)
        """

        gw_imshow(self._obj,
                  mask=mask,
                  nodata=nodata,
                  flip=flip,
                  text_color=text_color,
                  rot=rot,
                  **kwargs)

    def to_polygon(self, mask=None, connectivity=4):

        """
        Converts a ``dask`` array to a ``GeoDataFrame``

        Args:
            mask (Optional[numpy ndarray or rasterio Band object]): Must evaluate to bool (rasterio.bool_ or rasterio.uint8).
                Values of False or 0 will be excluded from feature generation. Note well that this is the inverse sense from
                Numpy's, where a mask value of True indicates invalid data in an array. If source is a Numpy masked array
                and mask is None, the source's mask will be inverted and used in place of mask.
            connectivity (Optional[int]): Use 4 or 8 pixel connectivity for grouping pixels into features.

        Returns:
            ``GeoDataFrame``

        Example:
            >>> import geowombat as gw
            >>>
            >>> with gw.open('image.tif') as src:
            >>>
            >>>     # Convert the input image to a GeoDataFrame
            >>>     df = src.gw.to_geodataframe(mask='source',
            >>>                                 num_workers=8)
        """

        return array_to_polygon(self._obj,
                                mask=mask,
                                connectivity=connectivity)

    def to_vector(self, filename, mask=None, connectivity=4):

        """
        Writes an Xarray DataArray to a vector file

        Args:
            filename (str): The output file name to write to.
            mask (numpy ndarray or rasterio Band object, optional): Must evaluate to bool (rasterio.bool_ or rasterio.uint8).
                Values of False or 0 will be excluded from feature generation. Note well that this is the inverse sense from
                Numpy's, where a mask value of True indicates invalid data in an array. If source is a Numpy masked array
                and mask is None, the source's mask will be inverted and used in place of mask.
            connectivity (Optional[int]): Use 4 or 8 pixel connectivity for grouping pixels into features.

        Returns:
            None
        """

        df_ = self.to_geodataframe(self._obj,
                                   mask=mask,
                                   connectivity=connectivity)

        df_.to_file(filename)

    def transform_crs(self,
                      dst_crs=None,
                      dst_res=None,
                      dst_width=None,
                      dst_height=None,
                      dst_bounds=None,
                      resampling='nearest',
                      warp_mem_limit=512,
                      num_threads=1):

        """
        Transforms a DataArray to a new coordinate reference system

        Args:
            dst_crs (Optional[CRS | int | dict | str]): The destination CRS.
            dst_res (Optional[tuple]): The destination resolution.
            dst_width (Optional[int]): The destination width. Cannot be used with ``dst_res``.
            dst_height (Optional[int]): The destination height. Cannot be used with ``dst_res``.
            dst_bounds (Optional[BoundingBox | tuple]): The destination bounds, as a ``rasterio.coords.BoundingBox``
                or as a tuple of (left, bottom, right, top).
            resampling (Optional[str]): The resampling method if ``filename`` is a ``list``.
                Choices are ['average', 'bilinear', 'cubic', 'cubic_spline', 'gauss', 'lanczos', 'max', 'med', 'min', 'mode', 'nearest'].
            warp_mem_limit (Optional[int]): The warp memory limit.
            num_threads (Optional[int]): The number of parallel threads.

        Returns:
            ``xarray.DataArray``

        Example:
            >>> import geowombat as gw
            >>>
            >>> with gw.open('image.tif') as src:
            >>>     dst = src.gw.transform_crs(4326)
        """

        return _transform_crs(self._obj,
                              dst_crs=dst_crs,
                              dst_res=dst_res,
                              dst_width=dst_width,
                              dst_height=dst_height,
                              dst_bounds=dst_bounds,
                              resampling=resampling,
                              warp_mem_limit=warp_mem_limit,
                              num_threads=num_threads)

    def to_crs(self,
               dst_crs=None,
               dst_res=None,
               dst_width=None,
               dst_height=None,
               dst_bounds=None,
               resampling='nearest',
               warp_mem_limit=512,
               num_threads=1):

        """
        Transforms a DataArray to a new coordinate reference system

        Args:
            dst_crs (Optional[CRS | int | dict | str]): The destination CRS.
            dst_res (Optional[tuple]): The destination resolution.
            dst_width (Optional[int]): The destination width. Cannot be used with ``dst_res``.
            dst_height (Optional[int]): The destination height. Cannot be used with ``dst_res``.
            dst_bounds (Optional[BoundingBox | tuple]): The destination bounds, as a ``rasterio.coords.BoundingBox``
                or as a tuple of (left, bottom, right, top).
            resampling (Optional[str]): The resampling method if ``filename`` is a ``list``.
                Choices are ['average', 'bilinear', 'cubic', 'cubic_spline', 'gauss', 'lanczos', 'max', 'med', 'min', 'mode', 'nearest'].
            warp_mem_limit (Optional[int]): The warp memory limit.
            num_threads (Optional[int]): The number of parallel threads.

        Returns:
            ``xarray.DataArray``

        Example:
            >>> import geowombat as gw
            >>>
            >>> with gw.open('image.tif') as src:
            >>>     dst = src.gw.to_crs(4326)
        """

        return _to_crs(self._obj,
                       dst_crs=dst_crs,
                       dst_res=dst_res,
                       dst_width=dst_width,
                       dst_height=dst_height,
                       dst_bounds=dst_bounds,
                       resampling=resampling,
                       warp_mem_limit=warp_mem_limit,
                       num_threads=num_threads)

    def to_raster(self,
                  filename,
                  readxsize=None,
                  readysize=None,
                  separate=False,
                  use_dask_store=False,
                  out_block_type='zarr',
                  keep_blocks=False,
                  verbose=0,
                  overwrite=False,
                  gdal_cache=512,
                  scheduler='processes',
                  n_jobs=1,
                  n_workers=None,
                  n_threads=None,
                  n_chunks=None,
                  overviews=False,
                  resampling='nearest',
                  use_client=False,
                  address=None,
                  total_memory=48,
                  driver='GTiff',
                  nodata=None,
                  blockxsize=512,
                  blockysize=512,
                  tags=None,
                  **kwargs):

        """
        Writes an Xarray DataArray to a raster file

        Args:
            filename (str): The output file name to write to.
            readxsize (Optional[int]): The size of column chunks to read. If not given, ``readxsize`` defaults to Dask chunk size.
            readysize (Optional[int]): The size of row chunks to read. If not given, ``readysize`` defaults to Dask chunk size.
            separate (Optional[bool]): Whether to write blocks as separate files. Otherwise, write to a single file.
            use_dask_store (Optional[bool]): Whether to use ``dask.array.store`` to save with Dask task graphs.
            out_block_type (Optional[str]): The output block type. Choices are ['gtiff', 'zarr'].
                Only used if ``separate`` = ``True``.
            keep_blocks (Optional[bool]): Whether to keep the blocks stored on disk. Only used if ``separate`` = ``True``.
            verbose (Optional[int]): The verbosity level.
            overwrite (Optional[bool]): Whether to overwrite an existing file.
            gdal_cache (Optional[int]): The ``GDAL`` cache size (in MB).
            scheduler (Optional[str]): The ``concurrent.futures`` scheduler to use. Choices are ['processes', 'threads'].
            n_jobs (Optional[int]): The total number of parallel jobs.
            n_workers (Optional[int]): The number of processes.
            n_threads (Optional[int]): The number of threads.
            n_chunks (Optional[int]): The chunk size of windows. If not given, equal to ``n_workers`` x 3.
            overviews (Optional[bool or list]): Whether to build overview layers.
            resampling (Optional[str]): The resampling method for overviews when ``overviews`` is ``True`` or a ``list``.
                Choices are ['average', 'bilinear', 'cubic', 'cubic_spline', 'gauss', 'lanczos', 'max', 'med', 'min', 'mode', 'nearest'].
            use_client (Optional[bool]): Whether to use a ``dask`` client.
            address (Optional[str]): A cluster address to pass to client. Only used when ``use_client`` = ``True``.
            total_memory (Optional[int]): The total memory (in GB) required when ``use_client`` = ``True``.
            driver (Optional[str]): The raster driver.
            nodata (Optional[int]): A 'no data' value.
            blockxsize (Optional[int]): The output x block size. Ignored if ``separate`` = ``True``.
            blockysize (Optional[int]): The output y block size. Ignored if ``separate`` = ``True``.
            tags (Optional[dict]): Image tags to write to file.
            kwargs (Optional[dict]): Additional keyword arguments to pass to ``rasterio.write``.

        Returns:
            None

        Examples:
            >>> import geowombat as gw
            >>>
            >>> # Use dask.compute()
            >>> with gw.open('input.tif') as ds:
            >>>     ds.gw.to_raster('output.tif', n_jobs=8)
            >>>
            >>> # Use a dask client
            >>> with gw.open('input.tif') as ds:
            >>>     ds.gw.to_raster('output.tif', use_client=True, n_workers=8, n_threads=4)
            >>>
            >>> # Compress the output
            >>> with gw.open('input.tif') as ds:
            >>>     ds.gw.to_raster('output.tif', n_jobs=8, compress='lzw')
        """

        if not hasattr(self._obj, 'crs'):
            raise AttributeError('The DataArray does not have a `crs` attribute.')

        if not hasattr(self._obj, 'transform'):
            raise AttributeError('The DataArray does not have a `transform` attribute.')

        kwargs = self._update_kwargs(nodata=nodata,
                                     driver=driver,
                                     blockxsize=blockxsize,
                                     blockysize=blockysize,
                                     **kwargs)

        to_raster(self._obj,
                  filename,
                  readxsize=readxsize,
                  readysize=readysize,
                  use_dask_store=use_dask_store,
                  separate=separate,
                  out_block_type=out_block_type,
                  keep_blocks=keep_blocks,
                  verbose=verbose,
                  overwrite=overwrite,
                  gdal_cache=gdal_cache,
                  scheduler=scheduler,
                  n_jobs=n_jobs,
                  n_workers=n_workers,
                  n_threads=n_threads,
                  n_chunks=n_chunks,
                  overviews=overviews,
                  resampling=resampling,
                  use_client=use_client,
                  address=address,
                  total_memory=total_memory,
                  crs=self._obj.crs,
                  transform=self._obj.transform,
                  width=self._obj.gw.ncols,
                  height=self._obj.gw.nrows,
                  count=self._obj.gw.nbands,
                  dtype=self._obj.data.dtype.name,
                  tags=tags,
                  **kwargs)

    def to_vrt(self,
               filename,
               resampling=None,
               nodata=None,
               init_dest_nodata=True,
               warp_mem_limit=128):

        """
        Writes a file to a VRT file

        Args:
            data (DataArray): The ``xarray.DataArray`` to write.
            filename (str): The output file name to write to.
            resampling (Optional[object]): The resampling algorithm for ``rasterio.vrt.WarpedVRT``.
            nodata (Optional[float or int]): The 'no data' value for ``rasterio.vrt.WarpedVRT``.
            init_dest_nodata (Optional[bool]): Whether or not to initialize output to ``nodata`` for ``rasterio.vrt.WarpedVRT``.
            warp_mem_limit (Optional[int]): The GDAL memory limit for ``rasterio.vrt.WarpedVRT``.

        Example:
            >>> import geowombat as gw
            >>> from rasterio.enums import Resampling
            >>>
            >>> with gw.config.update(ref_crs=102033):
            >>>
            >>>     with gw.open('image.tif') as ds:
            >>>
            >>>         ds.gw.to_vrt('image.vrt',
            >>>                      resampling=Resampling.cubic,
            >>>                      warp_mem_limit=256)
        """

        to_vrt(self._obj,
               filename,
               resampling=resampling,
               nodata=nodata,
               init_dest_nodata=init_dest_nodata,
               warp_mem_limit=warp_mem_limit)

    def predict(self,
                clf,
                outname=None,
                chunksize='same',
                x_chunks=(5000, 1),
                overwrite=False,
                return_as='array',
                n_jobs=1,
                backend='dask',
                verbose=0,
                nodata=None,
                dtype='uint8',
                gdal_cache=512,
                **kwargs):

        """
        Predicts an image using a pre-fit model

        Args:
            clf (object): A fitted classifier ``geowombat.model.Model`` instance with a ``predict`` method.
            outname (Optional[str]): An file name for the predictions.
            chunksize (Optional[str or tuple]): The chunk size for I/O. Default is 'same', or use the input chunk size.
            x_chunks (Optional[tuple]): The chunk size for the X predictors (or ``data``).
            overwrite (Optional[bool]): Whether to overwrite an existing file.
            return_as (Optional[str]): Whether to return the predictions as a ``xarray.DataArray`` or ``xarray.Dataset``.
                Only relevant if ``outname`` is not given.
            nodata (Optional[int or float]): The 'no data' value in the predictors.
            n_jobs (Optional[int]): The number of parallel jobs (chunks) for writing.
            backend (Optional[str]): The ``joblib`` backend scheduler.
            verbose (Optional[int]): The verbosity level.
            dtype (Optional[str]): The output data type passed to ``rasterio.write``.
            gdal_cache (Optional[int]): The GDAL cache (in MB) passed to ``rasterio.write``.
            kwargs (Optional[dict]): Additional keyword arguments passed to ``rasterio.write``.
                The ``blockxsize`` and ``blockysize`` should be excluded because they are taken from ``chunksize``.

        Returns:
            ``xarray.DataArray``

        Examples:
            >>> import geowombat as gw
            >>> from sklearn import ensemble
            >>>
            >>> clf = ensemble.RandomForestClassifier()
            >>> clf.fit(X, y)
            >>>
            >>> with gw.open('image.tif') as ds:
            >>>     pred = ds.gw.predict(clf)
        """

        kwargs = self._update_kwargs(**kwargs)

        return predict(self._obj,
                       clf,
                       outname=outname,
                       chunksize=chunksize,
                       x_chunks=x_chunks,
                       overwrite=overwrite,
                       return_as=return_as,
                       n_jobs=n_jobs,
                       backend=backend,
                       verbose=verbose,
                       nodata=nodata,
                       dtype=dtype,
                       gdal_cache=gdal_cache,
                       **kwargs)

    def apply(self, filename, user_func, n_jobs=1, **kwargs):

        """
        Applies a user function to an Xarray Dataset or DataArray and writes to file

        Args:
            filename (str): The output file name to write to.
            user_func (func): The user function to apply.
            n_jobs (Optional[int]): The number of parallel jobs for the cluster.
            kwargs (Optional[dict]): Keyword arguments passed to `to_raster`.

        Example:
            >>> import geowombat as gw
            >>> import xarray as xr
            >>>
            >>> def user_func(ds_):
            >>>     return ds_.max(axis=0)
            >>>
            >>> with xr.open_rasterio('image.tif', chunks=(1, 512, 512)) as ds:
            >>>     ds.io.apply('output.tif', user_func, n_jobs=8, overwrite=True, blockxsize=512, blockysize=512)
        """

        cluster = _Cluster(n_workers=n_jobs,
                           threads_per_worker=1,
                           scheduler_port=0,
                           processes=False)

        cluster.start()

        with joblib.parallel_backend('loky', n_jobs=n_jobs):

            ds_sub = user_func(self._obj)
            ds_sub.attrs = self._obj.attrs
            ds_sub.io.to_raster(filename, n_jobs=n_jobs, **kwargs)

        cluster.stop()

    def clip(self, df, query=None, mask_data=False):

        """
        Clips a DataArray

        Args:
            df (GeoDataFrame): The ``geopandas.GeoDataFrame`` to clip to.
            query (Optional[str]): A query to apply to ``df``.
            mask_data (Optional[bool]): Whether to mask values outside of the ``df`` geometry envelope.

        Returns:
             ``xarray.DataArray``
        """

        return clip(self._obj, df, query=query, mask_data=mask_data)

    def mask(self, df, query=None, keep='in'):

        """
        Masks a DataArray

        Args:
            df (GeoDataFrame or str): The ``geopandas.GeoDataFrame`` or filename to use for masking.
            query (Optional[str]): A query to apply to ``df``.
            keep (Optional[str]): If ``keep`` = 'in', mask values outside of the geometry (keep inside).
                Otherwise, if ``keep`` = 'out', mask values inside (keep outside).

        Returns:
             ``xarray.DataArray``
        """

        return mask(self._obj, df, query=query, keep=keep)

    def subset(self,
               left=None,
               top=None,
               right=None,
               bottom=None,
               rows=None,
               cols=None,
               center=False,
               mask_corners=False):

        """
        Subsets a DataArray

        Args:
            left (Optional[float]): The left coordinate.
            top (Optional[float]): The top coordinate.
            right (Optional[float]): The right coordinate.
            bottom (Optional[float]): The bottom coordinate.
            rows (Optional[int]): The number of output rows.
            cols (Optional[int]): The number of output rows.
            center (Optional[bool]): Whether to center the subset on ``left`` and ``top``.
            mask_corners (Optional[bool]): Whether to mask corners (requires ``pymorph``).
            chunksize (Optional[tuple]): A new chunk size for the output.

        Returns:
            ``xarray.DataArray``

        Example:
            >>> geowombat as gw
            >>>
            >>> with gw.open('image.tif', chunks=(1, 512, 512)) as ds:
            >>>     ds_sub = ds.gw.subset(-263529.884, 953985.314, rows=2048, cols=2048)
        """

        return subset(self._obj,
                      left=left,
                      top=top,
                      right=right,
                      bottom=bottom,
                      rows=rows,
                      cols=cols,
                      center=center,
                      mask_corners=mask_corners)

    def sample(self,
               method='random',
               band=None,
               n=None,
               strata=None,
               spacing=None,
               min_dist=None,
               max_attempts=10,
               **kwargs):

        """
        Generates samples from a raster

        Args:
            data (DataArray): The ``xarray.DataArray`` to extract data from.
            method (Optional[str]): The sampling method. Choices are ['random', 'systematic'].
            band (Optional[int or str]): The band name to extract from. Only required if ``method`` = 'random' and ``strata`` is given.
            n (Optional[int]): The total number of samples. Only required if ``method`` = 'random'.
            strata (Optional[dict]): The strata to sample within. The dictionary key-->value pairs should be {'conditional,value': proportion}.

                E.g.,

                    strata = {'==,1': 0.5, '>=,2': 0.5}
                    ... would sample 50% of total samples within class 1 and 50% of total samples in class >= 2.

                    strata = {'==,1': 10, '>=,2': 20}
                    ... would sample 10 samples within class 1 and 20 samples in class >= 2.

            spacing (Optional[float]): The spacing (in map projection units) when ``method`` = 'systematic'.
            min_dist (Optional[float or int]): A minimum distance allowed between samples. Only applies when ``method`` = 'random'.
            max_attempts (Optional[int]): The maximum numer of attempts to sample points > ``min_dist`` from each other.
            kwargs (Optional[dict]): Keyword arguments passed to ``geowombat.extract``.

        Returns:
            ``geopandas.GeoDataFrame``

        Examples:
            >>> import geowombat as gw
            >>>
            >>> # Sample 100 points randomly across the image
            >>> with gw.open('image.tif') as ds:
            >>>     df = ds.gw.sample(n=100)
            >>>
            >>> # Sample points systematically (with 10km spacing) across the image
            >>> with gw.open('image.tif') as ds:
            >>>     df = ds.gw.sample(method='systematic', spacing=10000.0)
            >>>
            >>> # Sample 50% of 100 in class 1 and 50% in classes >= 2
            >>> strata = {'==,1': 0.5, '>=,2': 0.5}
            >>> with gw.open('image.tif') as ds:
            >>>     df = ds.gw.sample(band=1, n=100, strata=strata)
            >>>
            >>> # Specify a per-stratum minimum allowed point distance of 1,000 meters
            >>> with gw.open('image.tif') as ds:
            >>>     df = ds.gw.sample(band=1, n=100, min_dist=1000, strata=strata)
        """

        return extract(self._obj,
                       method=method,
                       band=band,
                       n=n,
                       strata=strata,
                       spacing=spacing,
                       min_dist=min_dist,
                       max_attempts=max_attempts,
                       **kwargs)

    def extract(self,
                aoi,
                bands=None,
                time_names=None,
                band_names=None,
                frac=1.0,
                all_touched=False,
                mask=None,
                n_jobs=8,
                verbose=0,
                **kwargs):

        """
        Extracts data within an area or points of interest. Projections do not
        need to match, as they are handled 'on-the-fly'.

        Args:
            aoi (str or GeoDataFrame): A file or ``geopandas.GeoDataFrame`` to extract data frame.
            bands (Optional[int or 1d array-like]): A band or list of bands to extract.
                If not given, all bands are used. Bands should be GDAL-indexed (i.e., the first band is 1, not 0).
            band_names (Optional[list]): A list of band names. Length should be the same as `bands`.
            time_names (Optional[list]): A list of time names.
            frac (Optional[float]): A fractional subset of points to extract in each polygon feature.
            all_touched (Optional[bool]): The ``all_touched`` argument is passed to ``rasterio.features.rasterize``.
            mask (Optional[GeoDataFrame or Shapely Polygon]): A ``shapely.geometry.Polygon`` mask to subset to.
            n_jobs (Optional[int]): The number of features to rasterize in parallel.
            verbose (Optional[int]): The verbosity level.
            kwargs (Optional[dict]): Keyword arguments passed to ``dask.compute``.

        Returns:
            ``geopandas.GeoDataFrame``

        Examples:
            >>> import geowombat as gw
            >>>
            >>> with gw.open('image.tif') as ds:
            >>>     df = ds.gw.extract('poly.gpkg')
        """

        kwargs = self._update_kwargs(**kwargs)

        return extract(self._obj,
                       aoi,
                       bands=bands,
                       time_names=time_names,
                       band_names=band_names,
                       frac=frac,
                       all_touched=all_touched,
                       mask=mask,
                       n_jobs=n_jobs,
                       verbose=verbose,
                       **kwargs)

    def moving(self,
               band_coords='band',
               stat='mean',
               perc=50,
               nodata=None,
               w=3,
               n_jobs=1):

        """
        Applies a moving window function to the DataArray

        Args:
            band_coords (Optional[str]): The band coordinate name.
            stat (Optional[str]): The statistic to compute. Choices are ['mean', 'std', 'var', 'min', 'max', 'perc'].
            perc (Optional[int]): The percentile to return if ``stat`` = 'perc'.
            nodata (Optional[int or float]): A 'no data' value to ignore.
            w (Optional[int]): The moving window size (in pixels).
            n_jobs (Optional[int]): The number of rows to process in parallel.

        Returns:
            ``xarray.DataArray``

        Examples:
            >>> import geowombat as gw
            >>>
            >>> # Calculate the mean within a 5x5 window
            >>> with gw.open('image.tif') as src:
            >>>     res = src.gw.moving(stat='mean', w=5, nodata=32767.0, n_jobs=8)
            >>>
            >>> # Calculate the 90th percentile within a 15x15 window
            >>> with gw.open('image.tif') as src:
            >>>     res = src.gw.moving(stat='perc', w=15, perc=90, nodata=32767.0, n_jobs=8)
        """

        return moving(self._obj,
                      band_names=self._obj.coords[band_coords].values,
                      perc=perc,
                      nodata=nodata,
                      w=w,
                      stat=stat,
                      n_jobs=n_jobs)

    def norm_diff(self, b1, b2, nodata=None, mask=False, sensor=None, scale_factor=1.0):

        r"""
        Calculates the normalized difference band ratio

        Args:
            data (DataArray): The ``xarray.DataArray`` to process.
            b1 (str): The band name of the first band.
            b2 (str): The band name of the second band.
            sensor (Optional[str]): sensor (Optional[str]): The data's sensor.
            nodata (Optional[int or float]): A 'no data' value to fill NAs with.
            mask (Optional[bool]): Whether to mask the results.
            scale_factor (Optional[float]): A scale factor to apply to the data.

        Equation:

            .. math::
                {norm}_{diff} = \frac{b2 - b1}{b2 + b1}

        Returns:
            ``xarray.DataArray``
        """

        return gw_norm_diff(self._obj, b1, b2, sensor=sensor, nodata=nodata, mask=mask, scale_factor=scale_factor)

    def evi(self, nodata=None, mask=False, sensor=None, scale_factor=1.0):

        r"""
        Calculates the enhanced vegetation index

        Args:
            data (DataArray): The ``xarray.DataArray`` to process.
            nodata (Optional[int or float]): A 'no data' value to fill NAs with.
            mask (Optional[bool]): Whether to mask the results.
            sensor (Optional[str]): The data's sensor.
            scale_factor (Optional[float]): A scale factor to apply to the data.

        Equation:

            .. math::

                EVI = 2.5 \times \frac{NIR - red}{NIR \times 6 \times red - 7.5 \times blue + 1}

        Returns:
            ``xarray.DataArray``
        """

        return gw_evi(self._obj, nodata=nodata, mask=mask, sensor=sensor, scale_factor=scale_factor)

    def evi2(self, nodata=None, mask=False, sensor=None, scale_factor=1.0):

        r"""
        Calculates the two-band modified enhanced vegetation index

        Args:
            data (DataArray): The ``xarray.DataArray`` to process.
            nodata (Optional[int or float]): A 'no data' value to fill NAs with.
            mask (Optional[bool]): Whether to mask the results.
            sensor (Optional[str]): The data's sensor.
            scale_factor (Optional[float]): A scale factor to apply to the data.

        Equation:

            .. math::

                EVI2 = 2.5 \times \frac{NIR - red}{NIR + 1 + 2.4 \times red}

        Returns:
            ``xarray.DataArray``
        """

        return gw_evi2(self._obj, nodata=nodata, mask=mask, sensor=sensor, scale_factor=scale_factor)

    def nbr(self, nodata=None, mask=False, sensor=None, scale_factor=1.0):

        r"""
        Calculates the normalized burn ratio

        Args:
            data (DataArray): The ``xarray.DataArray`` to process.
            nodata (Optional[int or float]): A 'no data' value to fill NAs with.
            mask (Optional[bool]): Whether to mask the results.
            sensor (Optional[str]): The data's sensor.
            scale_factor (Optional[float]): A scale factor to apply to the data.

        Equation:

            .. math::
                NBR = \frac{NIR - SWIR1}{NIR + SWIR1}

        Returns:
            ``xarray.DataArray``
        """

        return gw_nbr(self._obj, nodata=nodata, mask=mask, sensor=sensor, scale_factor=scale_factor)

    def ndvi(self, nodata=None, mask=False, sensor=None, scale_factor=1.0):

        r"""
        Calculates the normalized difference vegetation index

        Args:
            data (DataArray): The ``xarray.DataArray`` to process.
            nodata (Optional[int or float]): A 'no data' value to fill NAs with.
            mask (Optional[bool]): Whether to mask the results.
            sensor (Optional[str]): The data's sensor.
            scale_factor (Optional[float]): A scale factor to apply to the data.

        Equation:

            .. math::
                NDVI = \frac{NIR - red}{NIR + red}

        Returns:
            ``xarray.DataArray``
        """

        return gw_ndvi(self._obj, nodata=nodata, mask=mask, sensor=sensor, scale_factor=scale_factor)

    def wi(self, nodata=None, mask=False, sensor=None, scale_factor=1.0):

        r"""
        Calculates the woody vegetation index

        Args:
            data (DataArray): The ``xarray.DataArray`` to process.
            nodata (Optional[int or float]): A 'no data' value to fill NAs with.
            mask (Optional[bool]): Whether to mask the results.
            sensor (Optional[str]): The data's sensor.
            scale_factor (Optional[float]): A scale factor to apply to the data.

        Equation:

            .. math::
                WI = SWIR1 + red

        Returns:
            ``xarray.DataArray``
        """

        return gw_wi(self._obj, nodata=nodata, mask=mask, sensor=sensor, scale_factor=scale_factor)

    def tasseled_cap(self, nodata=None, sensor=None, scale_factor=1.0):

        r"""
        Applies a tasseled cap transformation

        Args:
            nodata (Optional[int or float]): A 'no data' value to fill NAs with.
            sensor (Optional[str]): The data's sensor.
            scale_factor (Optional[float]): A scale factor to apply to the data.

        Returns:
            ``xarray.DataArray``

        Examples:
            >>> import geowombat as gw
            >>>
            >>> with gw.config.update(sensor='qb', scale_factor=0.0001):
            >>>     with gw.open('image.tif', band_names=['blue', 'green', 'red', 'nir']) as ds:
            >>>         tcap = ds.gw.tasseled_cap()
        """

        return gw_tasseled_cap(self._obj, nodata=nodata, sensor=sensor, scale_factor=scale_factor)

    def norm_brdf(self,
                  solar_za,
                  solar_az,
                  sensor_za,
                  sensor_az,
                  sensor=None,
                  wavelengths=None,
                  nodata=None,
                  mask=None,
                  scale_factor=1.0,
                  scale_angles=True):

        """
        Applies Bidirectional Reflectance Distribution Function (BRDF) normalization

        Args:
            solar_za (2d DataArray): The solar zenith angles (degrees).
            solar_az (2d DataArray): The solar azimuth angles (degrees).
            sensor_za (2d DataArray): The sensor azimuth angles (degrees).
            sensor_az (2d DataArray): The sensor azimuth angles (degrees).
            sensor (Optional[str]): The satellite sensor.
            wavelengths (str list): The wavelength(s) to normalize.
            nodata (Optional[int or float]): A 'no data' value to fill NAs with.
            mask (Optional[DataArray]): A data mask, where clear values are 0.
            scale_factor (Optional[float]): A scale factor to apply to the input data.
            scale_angles (Optional[bool]): Whether to scale the pixel angle arrays.

        Returns:
            ``xarray.DataArray``

        Examples:
            >>> import geowombat as gw
            >>>
            >>> # Example where pixel angles are stored in separate GeoTiff files
            >>> with gw.config.update(sensor='l7', scale_factor=0.0001, nodata=0):
            >>>
            >>>     with gw.open('solarz.tif') as solarz,
            >>>         gw.open('solara.tif') as solara,
            >>>             gw.open('sensorz.tif') as sensorz,
            >>>                 gw.open('sensora.tif') as sensora:
            >>>
            >>>         with gw.open('landsat.tif') as ds:
            >>>             ds_brdf = ds.gw.norm_brdf(solarz, solara, sensorz, sensora)
        """

        # Get the central latitude
        central_lat = project_coords(np.array([self._obj.x.values[int(self._obj.x.shape[0] / 2)]], dtype='float64'),
                                     np.array([self._obj.y.values[int(self._obj.y.shape[0] / 2)]], dtype='float64'),
                                     self._obj.crs,
                                     {'init': 'epsg:4326'})[1][0]

        return _BRDF().norm_brdf(self._obj,
                                 solar_za,
                                 solar_az,
                                 sensor_za,
                                 sensor_az,
                                 central_lat,
                                 sensor=sensor,
                                 wavelengths=wavelengths,
                                 nodata=nodata,
                                 mask=mask,
                                 scale_factor=scale_factor,
                                 scale_angles=scale_angles)



from pathlib import Path
from collections import namedtuple

import geopandas as gpd
from rasterio.coords import BoundingBox
from affine import Affine
import shapely
from shapely.geometry import Polygon

shapely.speedups.enable()


WavelengthsPan = namedtuple('WavelengthsPan', 'pan')
WavelengthsBGR = namedtuple('WavelengthsBGR', 'blue green red')
WavelengthsRGB = namedtuple('WavelengthsRGB', 'red green blue')
WavelengthsBGRN = namedtuple('WavelengthsBGRN', 'blue green red nir')
WavelengthsRGBN = namedtuple('WavelengthsRGBN', 'red green blue nir')
WavelengthsL57 = namedtuple('WavelengthsL57', 'blue green red nir swir1 swir2')
WavelengthsL57Thermal = namedtuple('WavelengthsL57Thermal', 'blue green red nir swir1 thermal swir2')
WavelengthsL57Pan = namedtuple('WavelengthsL57Pan', 'blue green red nir swir1 swir2 pan')
WavelengthsL8 = namedtuple('WavelengthsL8', 'coastal blue green red nir swir1 swir2 cirrus')
WavelengthsL8Thermal = namedtuple('WavelengthsL8Thermal', 'coastal blue green red nir swir1 swir2 cirrus tirs1 tirs2')
WavelengthsS2 = namedtuple('WavelengthsS2', 'blue green red nir1 nir2 nir3 nir rededge swir1 swir2')
WavelengthsS2Full = namedtuple('WavelengthsS2', 'coastal blue green red nir1 nir2 nir3 nir rededge water cirrus swir1 swir2')
WavelengthsS220 = namedtuple('WavelengthsS220', 'nir1 nir2 nir3 rededge swir1 swir2')


class DataProperties(object):

    @property
    def avail_sensors(self):
        """Get supported sensors"""
        return sorted(list(self.wavelengths.keys()))

    @property
    def altitude(self):

        """
        Get satellite altitudes (in km)
        """

        return dict(aster=705,
                    l5=705,
                    l7=705,
                    l8=705,
                    s2=786,
                    ps=475,
                    qb=482,
                    ik=681,
                    wv3=617)

    @property
    def central_um(self):

        """
        Get a dictionary of central wavelengths (in micrometers)
        """

        return dict(l5=WavelengthsL57(blue=0.485,
                                      green=0.56,
                                      red=0.66,
                                      nir=0.835,
                                      swir1=1.65,
                                      swir2=2.22),
                    l7=WavelengthsL57(blue=0.485,
                                      green=0.56,
                                      red=0.66,
                                      nir=0.835,
                                      swir1=1.65,
                                      swir2=2.22),
                    l7th=WavelengthsL57Thermal(blue=0.485,
                                               green=0.56,
                                               red=0.66,
                                               nir=0.835,
                                               swir1=1.65,
                                               thermal=11.45,
                                               swir2=2.22),
                    l7mspan=WavelengthsL57Pan(blue=0.485,
                                              green=0.56,
                                              red=0.66,
                                              nir=0.835,
                                              swir1=1.65,
                                              swir2=2.22,
                                              pan=0.71),
                    l7pan=WavelengthsPan(pan=0.71),
                    l8=WavelengthsL8(coastal=0.44,
                                     blue=0.48,
                                     green=0.56,
                                     red=0.655,
                                     nir=0.865,
                                     swir1=1.61,
                                     swir2=2.2,
                                     cirrus=1.37),
                    l8l7=WavelengthsL57(blue=0.48,
                                        green=0.56,
                                        red=0.655,
                                        nir=0.865,
                                        swir1=1.61,
                                        swir2=2.2),
                    l8l7mspan=WavelengthsL57Pan(blue=0.48,
                                                green=0.56,
                                                red=0.655,
                                                nir=0.865,
                                                swir1=1.61,
                                                swir2=2.2,
                                                pan=0.59),
                    l8pan=WavelengthsPan(pan=0.59),
                    l5bgrn=WavelengthsBGRN(blue=0.485,
                                           green=0.56,
                                           red=0.66,
                                           nir=0.835),
                    l7bgrn=WavelengthsBGRN(blue=0.485,
                                           green=0.56,
                                           red=0.66,
                                           nir=0.835),
                    l8bgrn=WavelengthsBGRN(blue=0.48,
                                           green=0.56,
                                           red=0.655,
                                           nir=0.865),
                    s2=WavelengthsS2(blue=0.49,
                                     green=0.56,
                                     red=0.665,
                                     nir1=0.705,
                                     nir2=0.74,
                                     nir3=0.783,
                                     nir=0.842,
                                     rededge=0.865,
                                     swir1=1.61,
                                     swir2=2.19),
                    s2f=WavelengthsS2Full(coastal=0.443,
                                          blue=0.49,
                                          green=0.56,
                                          red=0.665,
                                          nir1=0.705,
                                          nir2=0.74,
                                          nir3=0.783,
                                          nir=0.842,
                                          rededge=0.865,
                                          water=0.945,
                                          cirrus=1.375,
                                          swir1=1.61,
                                          swir2=2.19),
                    s2l7=WavelengthsL57(blue=0.49,
                                        green=0.56,
                                        red=0.665,
                                        nir=0.842,
                                        swir1=1.61,
                                        swir2=2.19),
                    s210=WavelengthsBGRN(blue=0.49,
                                         green=0.56,
                                         red=0.665,
                                         nir=0.842),
                    s220=WavelengthsS220(nir1=0.705,
                                         nir2=0.74,
                                         nir3=0.783,
                                         rededge=0.865,
                                         swir1=1.61,
                                         swir2=2.19),
                    ps=WavelengthsBGRN(blue=0.485,
                                       green=0.545,
                                       red=0.63,
                                       nir=0.82))

    @property
    def sensor_names(self):

        """
        Get sensor full names
        """

        return dict(rgb='red, green, and blue',
                    rgbn='red, green, blue, and NIR',
                    bgr='blue, green, and red',
                    bgrn='blue, green, red, and NIR',
                    l5='Landsat 5 Thematic Mapper (TM)',
                    l7='Landsat 7 Enhanced Thematic Mapper Plus (ETM+) without panchromatic and thermal bands',
                    l7th='Landsat 7 Enhanced Thematic Mapper Plus (ETM+) with thermal band',
                    l7mspan='Landsat 7 Enhanced Thematic Mapper Plus (ETM+) with panchromatic band',
                    l7pan='Landsat 7 panchromatic band',
                    l8='Landsat 8 Operational Land Imager (OLI) and Thermal Infrared Sensor (TIRS) without panchromatic and thermal bands',
                    l8l7='Landsat 8 Operational Land Imager (OLI) and Thermal Infrared Sensor (TIRS) with 6 Landsat 7-like bands',
                    l8l7mspan='Landsat 8 Operational Land Imager (OLI) and panchromatic band with 6 Landsat 7-like bands',
                    l8th='Landsat 8 Operational Land Imager (OLI) and Thermal Infrared Sensor (TIRS) with thermal band',
                    l8pan='Landsat 8 panchromatic band',
                    s2='Sentinel 2 Multi-Spectral Instrument (MSI) without 3 60m bands (coastal, water vapor, cirrus)',
                    s2f='Sentinel 2 Multi-Spectral Instrument (MSI) with 3 60m bands (coastal, water vapor, cirrus)',
                    s2l7='Sentinel 2 Multi-Spectral Instrument (MSI) with 6 Landsat 7-like bands',
                    s210='Sentinel 2 Multi-Spectral Instrument (MSI) with 4 10m (visible + NIR) bands',
                    s220='Sentinel 2 Multi-Spectral Instrument (MSI) with 6 20m bands',
                    ps='PlanetScope with 4 (visible + NIR) bands',
                    qb='Quickbird with 4 (visible + NIR) bands',
                    ik='IKONOS with 4 (visible + NIR) bands')

    @property
    def wavelengths(self):

        """
        Get a dictionary of sensor wavelengths
        """

        return dict(rgb=WavelengthsRGB(red=1,
                                       green=2,
                                       blue=3),
                    rgbn=WavelengthsRGBN(red=1,
                                         green=2,
                                         blue=3,
                                         nir=4),
                    bgr=WavelengthsBGR(blue=1,
                                       green=2,
                                       red=3),
                    bgrn=WavelengthsBGRN(blue=1,
                                         green=2,
                                         red=3,
                                         nir=4),
                    l5=WavelengthsL57(blue=1,
                                      green=2,
                                      red=3,
                                      nir=4,
                                      swir1=5,
                                      swir2=6),
                    l7=WavelengthsL57(blue=1,
                                      green=2,
                                      red=3,
                                      nir=4,
                                      swir1=5,
                                      swir2=6),
                    l7th=WavelengthsL57Thermal(blue=1,
                                               green=2,
                                               red=3,
                                               nir=4,
                                               swir1=5,
                                               thermal=6,
                                               swir2=7),
                    l7mspan=WavelengthsL57Pan(blue=1,
                                              green=2,
                                              red=3,
                                              nir=4,
                                              swir1=5,
                                              swir2=6,
                                              pan=7),
                    l7pan=WavelengthsPan(pan=1),
                    l8=WavelengthsL8(coastal=1,
                                     blue=2,
                                     green=3,
                                     red=4,
                                     nir=5,
                                     swir1=6,
                                     swir2=7,
                                     cirrus=8),
                    l8th=WavelengthsL8Thermal(coastal=1,
                                              blue=2,
                                              green=3,
                                              red=4,
                                              nir=5,
                                              swir1=6,
                                              swir2=7,
                                              cirrus=8,
                                              tirs1=9,
                                              tirs2=10),
                    l8l7=WavelengthsL57(blue=1,
                                        green=2,
                                        red=3,
                                        nir=4,
                                        swir1=5,
                                        swir2=6),
                    l8l7mspan=WavelengthsL57Pan(blue=1,
                                                green=2,
                                                red=3,
                                                nir=4,
                                                swir1=5,
                                                swir2=6,
                                                pan=7),
                    l8pan=WavelengthsPan(pan=1),
                    l5bgrn=WavelengthsBGRN(blue=1,
                                           green=2,
                                           red=3,
                                           nir=4),
                    l7bgrn=WavelengthsBGRN(blue=1,
                                           green=2,
                                           red=3,
                                           nir=4),
                    l8bgrn=WavelengthsBGRN(blue=1,
                                           green=2,
                                           red=3,
                                           nir=4),
                    s2=WavelengthsS2(blue=1,
                                     green=2,
                                     red=3,
                                     nir1=4,
                                     nir2=5,
                                     nir3=6,
                                     nir=7,
                                     rededge=8,
                                     swir1=9,
                                     swir2=10),
                    s2f=WavelengthsS2Full(coastal=1,
                                          blue=2,
                                          green=3,
                                          red=4,
                                          nir1=5,
                                          nir2=6,
                                          nir3=7,
                                          nir=8,
                                          rededge=9,
                                          water=10,
                                          cirrus=11,
                                          swir1=12,
                                          swir2=13),
                    s2l7=WavelengthsL57(blue=1,
                                        green=2,
                                        red=3,
                                        nir=4,
                                        swir1=5,
                                        swir2=6),
                    s210=WavelengthsBGRN(blue=1,
                                         green=2,
                                         red=3,
                                         nir=4),
                    s220=WavelengthsS220(nir1=1,
                                         nir2=2,
                                         nir3=3,
                                         rededge=4,
                                         swir1=5,
                                         swir2=6),
                    ps=WavelengthsBGRN(blue=1,
                                       green=2,
                                       red=3,
                                       nir=4),
                    qb=WavelengthsBGRN(blue=1,
                                       green=2,
                                       red=3,
                                       nir=4),
                    ik=WavelengthsBGRN(blue=1,
                                       green=2,
                                       red=3,
                                       nir=4))

    @property
    def ndims(self):
        """Get the number of array dimensions"""
        return len(self._obj.shape)

    @property
    def row_chunks(self):
        """Get the row chunk size"""
        return self._obj.data.chunksize[-2]

    @property
    def col_chunks(self):
        """Get the column chunk size"""
        return self._obj.data.chunksize[-1]

    @property
    def band_chunks(self):

        """
        Get the band chunk size
        """

        if self.ndims > 2:
            return self._obj.data.chunksize[-3]
        else:
            return 1

    @property
    def time_chunks(self):

        """
        Get the time chunk size
        """

        if self.ndims > 3:
            return self._obj.data.chunksize[-4]
        else:
            return 1

    @property
    def ntime(self):

        """
        Get the number of time dimensions
        """

        if self.ndims > 3:
            return self._obj.shape[-4]
        else:
            return 1

    @property
    def nbands(self):

        """
        Get the number of array bands
        """

        if self.ndims > 2:
            return self._obj.shape[-3]
        else:
            return 1

    @property
    def nrows(self):
        """Get the number of array rows"""
        return self._obj.shape[-2]

    @property
    def ncols(self):
        """Get the number of array columns"""
        return self._obj.shape[-1]

    @property
    def left(self):

        """
        Get the array bounding box left coordinate

        Pixel shift reference:
            https://github.com/pydata/xarray/blob/master/xarray/backends/rasterio_.py
            http://web.archive.org/web/20160326194152/http://remotesensing.org/geotiff/spec/geotiff2.5.html#2.5.2
        """

        return float(self._obj.x.min().values) - self.cellxh

    @property
    def right(self):
        """Get the array bounding box right coordinate"""
        return float(self._obj.x.max().values) + self.cellxh

    @property
    def top(self):
        """Get the array bounding box top coordinate"""
        return float(self._obj.y.max().values) + self.cellyh

    @property
    def bottom(self):
        """Get the array bounding box bottom coordinate"""
        return float(self._obj.y.min().values) - self.cellyh

    @property
    def bounds(self):
        """Get the array bounding box (left, bottom, right, top)"""
        return self.left, self.bottom, self.right, self.top

    @property
    def bounds_as_namedtuple(self):
        """Get the array bounding box as a ``rasterio.coords.BoundingBox``"""
        return BoundingBox(left=self.left, bottom=self.bottom, right=self.right, top=self.top)

    @property
    def celly(self):
        """Get the cell size in the y direction"""
        return self._obj.res[1]

    @property
    def cellx(self):
        """Get the cell size in the x direction"""
        return self._obj.res[0]

    @property
    def cellyh(self):
        """Get the half width of the cell size in the y direction"""
        return self.celly / 2.0

    @property
    def cellxh(self):
        """Get the half width of the cell size in the x direction"""
        return self.cellx / 2.0

    @property
    def geometry(self):

        """
        Get the polygon geometry of the array bounding box
        """

        return Polygon([(self.left, self.bottom),
                        (self.left, self.top),
                        (self.right, self.top),
                        (self.right, self.bottom),
                        (self.left, self.bottom)])

    @property
    def geodataframe(self):

        """
        Get a ``geopandas.GeoDataFrame`` of the array bounds
        """

        return gpd.GeoDataFrame(data=[Path(self._obj.filename).name if hasattr(self._obj, 'filename') else 1],
                                columns=['grid'],
                                geometry=[self._obj.gw.geometry],
                                crs=self._obj.crs)

    @property
    def unary_union(self):
        """Get a representation of the union of the image bounds"""
        return shapely.ops.unary_union(self.geometry)

    @property
    def meta(self):

        """
        Get the array metdata
        """

        Meta = namedtuple('Meta', 'left right top bottom bounds affine geometry')

        return Meta(left=self.left,
                    right=self.right,
                    top=self.top,
                    bottom=self.bottom,
                    bounds=self.bounds,
                    affine=Affine(*self._obj.transform),
                    geometry=self.geometry)



from ..errors import logger

import xarray as xr
import matplotlib as mpl
import matplotlib.pyplot as plt


class Plotting(object):

    @staticmethod
    def imshow(data,
               mask=False,
               nodata=0,
               flip=False,
               text_color='black',
               rot=30,
               ax=None,
               **kwargs):

        """
        Shows an image on a plot

        Args:
            data (``xarray.DataArray`` or ``xarray.Dataset``): The data to plot.
            mask (Optional[bool]): Whether to mask 'no data' values (given by ``nodata``).
            nodata (Optional[int or float]): The 'no data' value.
            flip (Optional[bool]): Whether to flip an RGB array's band order.
            text_color (Optional[str]): The text color.
            rot (Optional[int]): The degree rotation for the x-axis tick labels.
            ax (Optional[object]): A ``matplotlib`` axis object.
            kwargs (Optional[dict]): Keyword arguments passed to ``xarray.plot.imshow``.

        Returns:
            ``matplotlib`` axis object

        Examples:
            >>> import geowombat as gw
            >>>
            >>> # Open a 3-band image and plot the first band
            >>> with gw.open('image.tif') as ds:
            >>>     ax = ds.sel(band=1).gw.imshow()
            >>>
            >>> # Open and plot a 3-band image
            >>> with gw.open('image.tif') as ds:
            >>>
            >>>     ax = ds.sel(band=['red', 'green', 'blue']).gw.imshow(mask=True,
            >>>                                                          nodata=0,
            >>>                                                          vmin=0.1,
            >>>                                                          vmax=0.9,
            >>>                                                          robust=True)
        """

        if data.gw.nbands != 1:

            if data.gw.nbands != 3:
                logger.exception('  Only 1-band or 3-band arrays can be plotted.')

        plt.rcParams['axes.titlesize'] = 5
        plt.rcParams['axes.titlepad'] = 5
        plt.rcParams['text.color'] = text_color
        plt.rcParams['axes.labelcolor'] = text_color
        plt.rcParams['xtick.color'] = text_color
        plt.rcParams['ytick.color'] = text_color
        plt.rcParams['figure.dpi'] = kwargs['dpi'] if 'dpi' in kwargs else 150
        plt.rcParams['savefig.bbox'] = 'tight'
        plt.rcParams['savefig.pad_inches'] = 0.5

        if not ax:
            fig, ax = plt.subplots()

        if mask:

            if isinstance(data, xr.Dataset):

                if data.gw.nbands == 1:
                    plot_data = data.where((data['mask'] < 3) & (data != nodata))
                else:
                    plot_data = data.where((data['mask'] < 3) & (data.max(axis=0) != nodata))

            else:

                if data.gw.nbands == 1:
                    plot_data = data.where(data != nodata)
                else:
                    plot_data = data.where(data.max(axis=0) != nodata)

        else:
            plot_data = data

        if plot_data.gw.nbands == 3:

            plot_data = plot_data.transpose('y', 'x', 'band')

            if flip:
                plot_data = plot_data[..., ::-1]

            plot_data.plot.imshow(rgb='band',
                                  ax=ax,
                                  **kwargs)

        else:

            plot_data.plot.imshow(ax=ax,
                                  **kwargs)

        ax.xaxis.set_major_formatter(mpl.ticker.StrMethodFormatter('{x:,.0f}'))
        ax.yaxis.set_major_formatter(mpl.ticker.StrMethodFormatter('{x:,.0f}'))

        for tick in ax.get_xticklabels():
            tick.set_rotation(rot)

        plt.show()

        return ax



from ..backends import to_crs
from ..backends import transform_crs
from .io import apply
from .io import to_raster
from .io import to_vrt
from .sops import SpatialOperations
from .conversion import Converters
from .util import MapProcesses
from .vi import VegetationIndices
from .vi import TasseledCap

# Imports intended for module level
from .util import sort_images_by_date

# Imports intended for package level
extract = SpatialOperations().extract
sample = SpatialOperations().sample
subset = SpatialOperations().subset
clip = SpatialOperations().clip
mask = SpatialOperations().mask
coregister = SpatialOperations().coregister
polygons_to_points = Converters().polygons_to_points
indices_to_coords = Converters().indices_to_coords
coords_to_indices = Converters().coords_to_indices
dask_to_xarray = Converters().dask_to_xarray
ndarray_to_xarray = Converters().ndarray_to_xarray
array_to_polygon = Converters().array_to_polygon
polygon_to_array = Converters().polygon_to_array
geodataframe_to_array = Converters().geodataframe_to_array
to_geodataframe = Converters().to_geodataframe
moving = MapProcesses().moving
norm_diff = VegetationIndices().norm_diff
evi = VegetationIndices().evi
evi2 = VegetationIndices().evi2
nbr = VegetationIndices().nbr
ndvi = VegetationIndices().ndvi
wi = VegetationIndices().wi
tasseled_cap = TasseledCap().tasseled_cap

__all__ = ['apply',
           'to_crs',
           'transform_crs',
           'to_raster',
           'to_vrt',
           'to_geodataframe',
           'geodataframe_to_array',
           'extract',
           'sample',
           'subset',
           'clip',
           'mask',
           'coregister',
           'polygons_to_points',
           'indices_to_coords',
           'coords_to_indices',
           'dask_to_xarray',
           'ndarray_to_xarray',
           'array_to_polygon',
           'polygon_to_array',
           'sort_images_by_date',
           'moving',
           'norm_diff',
           'evi',
           'evi2',
           'nbr',
           'ndvi',
           'wi',
           'tasseled_cap']
